import contextlib
from asyncio import TimeoutError
from collections import defaultdict
from typing import Optional, Callable, Any, Coroutine, cast, Iterable
from typing_extensions import Self

from interactions.api.events.base import RawGatewayEvent
from interactions.api.events.discord import MessageCreate
from interactions.api.events.internal import (
    CommandError,
    CommandCompletion,
    ExtensionUnload,
)
from interactions.client.client import Client
from interactions.client.utils.input_utils import get_args, get_first_word
from interactions.models.discord.enums import Intents
from interactions.models.discord.message import Message
from interactions.models.internal.listener import listen
from .command import PrefixedCommand
from .context import PrefixedContext
from .utils import when_mentioned

__all__ = ("PrefixedInjectedClient", "PrefixedManager", "setup")


class PrefixedInjectedClient(Client):
    """
    A semi-stub for a `Client` injected with prefixed commands.

    This should only be used for type hinting.
    """

    prefixed: "PrefixedManager"
    "The prefixed command manager for this client."


class PrefixedManager:
    """
    The main part of the extension. Deals with injecting itself in the first place.

    Parameters:
        client: The client instance.
        default_prefix: The default prefix to use. Defaults to `None`.
        generate_prefixes: An asynchronous function \
            that takes in a `Client` and `Message` object and returns either a \
            string or an iterable of strings. Defaults to `None`.
        prefixed_context: The object to instantiate for Prefixed Context

    """

    def __init__(
        self,
        client: Client,
        *,
        default_prefix: Optional[str | list[str]] = None,
        generate_prefixes: Optional[
            Callable[
                [Client, Message],
                Coroutine[Any, Any, str | list[str]],
            ]
        ] = None,
        prefixed_context: type[PrefixedContext] = PrefixedContext,
    ) -> None:
        # typehinting funkyness for better typehints
        client = cast(PrefixedInjectedClient, client)

        self.client = client
        self.default_prefix = default_prefix
        self.prefixed_context = prefixed_context
        self.commands: dict[str, PrefixedCommand] = {}
        self._ext_command_list: defaultdict[str, set[str]] = defaultdict(set)

        if (
            default_prefix or (generate_prefixes and generate_prefixes != when_mentioned)
        ) and Intents.MESSAGE_CONTENT not in client.intents:
            client.logger.warning(
                "Prefixed commands will not work since the required intent is not set -> Requires:"
                f" {Intents.MESSAGE_CONTENT.__repr__()} or usage of the default mention prefix as the prefix"
            )

        if default_prefix is None and generate_prefixes is None:
            # by default, use mentioning the bot as the prefix
            generate_prefixes = when_mentioned

        self.generate_prefixes = (  # type: ignore
            generate_prefixes if generate_prefixes is not None else self.generate_prefixes
        )

        self.client.prefixed = self

        self._dispatch_prefixed_commands = self._dispatch_prefixed_commands.copy_with_binding(self)
        self._handle_ext_unload = self._handle_ext_unload.copy_with_binding(self)

        self.client.add_listener(self._dispatch_prefixed_commands)
        self.client.add_listener(self._handle_ext_unload)

        self.client._add_command_hook.append(self._register_command)

    async def generate_prefixes(self, client: Client, msg: Message) -> str | list[str]:
        """
        Generates a list of prefixes a prefixed command can have based on the client and message.

        This can be overwritten by passing a function to generate_prefixes on initialization.

        Args:
            client: The client instance.
            msg: The message sent.

        Returns:
            The prefix(es) to check for.

        """
        return self.default_prefix  # type: ignore

    def add_command(self, command: PrefixedCommand) -> None:
        """
        Add a prefixed command to the manager.

        Args:
            command: The command to add.

        """
        if command.is_subcommand:
            raise ValueError("You cannot add subcommands to the client - add the base command instead.")

        command._parse_parameters()

        if self.commands.get(command.name):
            raise ValueError(f"Duplicate command! Multiple commands share the name/alias: {command.name}.")
        self.commands[command.name] = command

        for alias in command.aliases:
            if self.commands.get(alias):
                raise ValueError(f"Duplicate command! Multiple commands share the name/alias: {alias}.")
            self.commands[alias] = command

        if command.extension:
            self._ext_command_list[command.extension.extension_name].add(command.name)
            command.extension._commands.append(command)

    def get_command(self, name: str) -> Optional[PrefixedCommand]:
        """
        Gets a prefixed command by the name/alias specified.

        This function is able to resolve subcommands - fully qualified names can be used.
        For example, passing in ``foo bar`` would get the subcommand ``bar`` from the
        command ``foo``.

        Args:
            name: The name of the command to search for.

        Returns:
            The command object, if found.

        """
        if " " not in name:
            return self.commands.get(name)

        names = name.split()
        if not names:
            return None

        cmd = self.commands.get(names[0])
        if not cmd:
            return cmd

        for name in names[1:]:
            try:
                cmd = cmd.subcommands[name]
            except (AttributeError, KeyError):
                return None

        return cmd

    def _remove_cmd_and_aliases(self, name: str) -> None:
        if cmd := self.commands.pop(name, None):
            if cmd.extension:
                self._ext_command_list[cmd.extension.extension_name].discard(cmd.name)
                with contextlib.suppress(ValueError):
                    cmd.extension._commands.remove(cmd)

            for alias in cmd.aliases:
                self.commands.pop(alias, None)

    def remove_command(self, name: str, delete_parent_if_empty: bool = False) -> Optional[PrefixedCommand]:
        """
        Removes a prefixed command if it exists.

        If an alias is specified, only the alias will be removed.
        This function is able to resolve subcommands - fully qualified names can be used.
        For example, passing in ``foo bar`` would delete the subcommand ``bar``
        from the command ``foo``.

        Args:
            name: The command to remove.
            delete_parent_if_empty: Should the parent command be deleted if it \
                ends up having no subcommands after deleting the command specified? \
                Defaults to `False`.

        Returns:
            The command that was removed, if one was. If the command was not found,
            this function returns `None`.

        """
        command = self.get_command(name)

        if command is None:
            return None

        if name in command.aliases:
            command.aliases.remove(name)
            return command

        if command.parent:
            command.parent.remove_command(command.name)
        else:
            self._remove_cmd_and_aliases(command.name)

        if delete_parent_if_empty:
            while command.parent is not None and not command.parent.subcommands:
                if command.parent.parent:
                    _new_cmd = command.parent
                    command.parent.parent.remove_command(command.parent.name)
                    command = _new_cmd
                else:
                    self._remove_cmd_and_aliases(command.parent.name)
                    break

        return command

    def _register_command(self, callback: Callable) -> None:
        """Registers a prefixed command, if there is one given."""
        if not isinstance(callback, PrefixedCommand):
            return

        if not callback.is_subcommand:
            self.add_command(callback)

    @listen("extension_unload")
    async def _handle_ext_unload(self, event: ExtensionUnload) -> None:
        """Unregisters all prefixed commands in an extension as it is being unloaded."""
        for name in self._ext_command_list[event.extension.extension_name].copy():
            self.remove_command(name)

    @listen("raw_message_create", is_default_listener=True)
    async def _dispatch_prefixed_commands(self, event: RawGatewayEvent) -> None:  # noqa: C901
        """Determine if a prefixed command is being triggered, and dispatch it."""
        # don't waste time processing this if there are no prefixed commands
        if not self.commands:
            return

        data = event.data

        # many bots will not have the message content intent, and so will not have content
        # for most messages. since there's nothing for prefixed commands to work off of,
        # we might as well not waste time
        if not data.get("content"):
            return

        # webhooks and users labeled with the bot property are bots, and should be ignored
        if data.get("webhook_id") or data["author"].get("bot", False):
            return

        # now, we've done the basic filtering out, but everything from here on out relies
        # on a proper message object, so now we either hope its already in the cache or wait
        # on the processor

        # first, let's check the cache...
        message = self.client.cache.get_message(int(data["channel_id"]), int(data["id"]))

        # this huge if statement basically checks if the message hasn't been fully processed by
        # the processor yet, which would mean that these fields aren't fully filled
        if message and (
            (not message._guild_id and event.data.get("guild_id"))
            or (message._guild_id and not message.guild)
            or not message.channel
        ):
            message = None

        # if we didn't get a message, then we know we should wait for the message create event
        if not message:
            try:
                # i think 2 seconds is a very generous timeout limit
                msg_event: MessageCreate = await self.client.wait_for(
                    MessageCreate, checks=lambda e: int(e.message.id) == int(data["id"]), timeout=2
                )
                message = msg_event.message
            except TimeoutError:
                return

        if not message.content:
            return

        # here starts the actual prefixed command parsing part
        prefixes: str | Iterable[str] = await self.generate_prefixes(self.client, message)

        if isinstance(prefixes, str):
            # its easier to treat everything as if it may be an iterable
            # rather than building a special case for this
            prefixes = (prefixes,)  # type: ignore

        prefix_used = next(
            (prefix for prefix in prefixes if message.content.startswith(prefix)),
            None,
        )
        if not prefix_used:
            return

        context = self.prefixed_context.from_message(self.client, message)
        context.prefix = prefix_used

        content_parameters = message.content.removeprefix(prefix_used).strip()
        command: "Self | PrefixedCommand" = self  # yes, this is a hack

        while True:
            first_word: str = get_first_word(content_parameters)  # type: ignore
            if isinstance(command, PrefixedCommand):
                new_command = command.subcommands.get(first_word)
            else:
                new_command = command.commands.get(first_word)
            if not new_command or not new_command.enabled:
                break

            command = new_command
            content_parameters = content_parameters.removeprefix(first_word).strip()

        if not isinstance(command, PrefixedCommand) or not command.enabled:
            return

        context.command = command
        context.content_parameters = content_parameters.strip()
        context.args = get_args(context.content_parameters)
        try:
            if self.client.pre_run_callback:
                await self.client.pre_run_callback(context)
            await command(context)
            if self.client.post_run_callback:
                await self.client.post_run_callback(context)
        except Exception as e:
            self.client.dispatch(CommandError(ctx=context, error=e))
        finally:
            self.client.dispatch(CommandCompletion(ctx=context))


def setup(
    client: Client,
    *,
    default_prefix: Optional[str | list[str]] = None,
    generate_prefixes: Optional[
        Callable[
            [Client, Message],
            Coroutine[Any, Any, str | list[str]],
        ]
    ] = None,
    prefixed_context: type[PrefixedContext] = PrefixedContext,
) -> PrefixedManager:
    """
    Sets up prefixed commands. It is recommended to use this function directly to do so.

    Parameters:
        client: The client instance.
        default_prefix: The default prefix to use. Defaults to `None`.
        generate_prefixes: An asynchronous function \
            that takes in a `Client` and `Message` object and returns either a \
            string or an iterable of strings. Defaults to `None`.
        prefixed_context: The object to instantiate for Prefixed Context

    Returns:
        PrefixedManager: The class that deals with all things prefixed commands.

    """
    return PrefixedManager(
        client,
        default_prefix=default_prefix,
        generate_prefixes=generate_prefixes,
        prefixed_context=prefixed_context,
    )



import asyncio
import time
import typing
from enum import IntEnum
from typing import TYPE_CHECKING, Any, Dict, Type

if TYPE_CHECKING:
    from interactions.models.internal.context import BaseContext

__all__ = (
    "Buckets",
    "Cooldown",
    "CooldownSystem",
    "SlidingWindowSystem",
    "ExponentialBackoffSystem",
    "LeakyBucketSystem",
    "TokenBucketSystem",
    "MaxConcurrency",
)


class Buckets(IntEnum):
    """
    Outlines the cooldown buckets that may be used. Should a bucket for guilds exist, and the command is invoked in a DM, a sane default will be used.

    ??? note
         To add your own, override this

    """

    DEFAULT = 0
    """Default is the same as user"""
    USER = 1
    """Per user cooldowns"""
    GUILD = 2
    """Per guild cooldowns"""
    CHANNEL = 3
    """Per channel cooldowns"""
    MEMBER = 4
    """Per guild member cooldowns"""
    CATEGORY = 5
    """Per category cooldowns"""
    ROLE = 6
    """Per role cooldowns"""

    async def get_key(self, context: "BaseContext") -> Any:
        if self is Buckets.USER:
            return context.author.id
        if self is Buckets.GUILD:
            return context.guild_id or context.author.id
        if self is Buckets.CHANNEL:
            return context.channel.id
        if self is Buckets.MEMBER:
            return (context.guild_id, context.author.id) if context.guild_id else context.author.id
        if self is Buckets.CATEGORY:
            return await context.channel.parent_id if context.channel.parent else context.channel.id
        if self is Buckets.ROLE:
            return context.author.top_role.id if context.guild_id else context.channel.id
        return context.author.id

    def __call__(self, context: "BaseContext") -> Any:
        return self.get_key(context)


class CooldownSystem:
    """
    A basic cooldown strategy that allows a specific number of commands to be executed within a given interval. Once the rate is reached, no more tokens can be acquired until the interval has passed.

    Attributes:
        rate: The number of commands allowed per interval.
        interval: The time window (in seconds) within which the allowed number of commands can be executed.

    ??? tip "Example Use-case"
        This strategy is useful for scenarios where you want to limit the number of times a command can be executed within a fixed time frame, such as preventing command spamming or limiting API calls.

    """

    __slots__ = "rate", "interval", "opened", "_tokens"

    def __init__(self, rate: int, interval: float) -> None:
        self.rate: int = rate
        self.interval: float = interval
        self.opened: float = 0.0

        self._tokens: int = self.rate

        # sanity checks
        if self.rate == 0:
            raise ValueError("Cooldown rate must be greater than 0")
        if self.interval == 0:
            raise ValueError("Cooldown interval must be greater than 0")

    def reset(self) -> None:
        """Resets the tokens for this cooldown."""
        self._tokens = self.rate
        self.opened = 0.0

    def on_cooldown(self) -> bool:
        """
        Returns the cooldown state of the command.

        Returns:
            boolean state if the command is on cooldown or not

        """
        self.determine_cooldown()

        return self._tokens == 0

    def acquire_token(self) -> bool:
        """
        Attempt to acquire a token for a command to run.

        Returns:
            True if a token was acquired, False if not

        """
        self.determine_cooldown()

        if self._tokens == 0:
            return False
        if self._tokens == self.rate:
            self.opened = time.time()
        self._tokens -= 1

        return True

    def get_cooldown_time(self) -> float:
        """
        Returns how long until the cooldown will reset.

        Returns:
            remaining cooldown time, will return 0 if the cooldown has not been reached

        """
        self.determine_cooldown()
        return 0 if self._tokens != 0 else self.interval - (time.time() - self.opened)

    def determine_cooldown(self) -> None:
        """Determines the state of the cooldown system."""
        c_time = time.time()

        if c_time > self.opened + self.interval:
            # cooldown has expired, reset the cooldown
            self.reset()


class SlidingWindowSystem(CooldownSystem):
    """
    A sliding window cooldown strategy that allows a specific number of commands to be executed within a rolling time window.

    The cooldown incrementally resets as commands fall outside of the window.

    Attributes:
        rate: The number of commands allowed per interval.
        interval: The time window (in seconds) within which the allowed number of commands can be executed.

    ??? tip "Example Use-case"
        This strategy is useful for scenarios where you want to limit the rate of commands executed over a continuous time window, such as ensuring consistent usage of resources or controlling chat bot response frequency.

    """

    __slots__ = "rate", "interval", "timestamps"

    def __init__(self, rate: int, interval: float) -> None:
        self.rate: int = rate
        self.interval: float = interval
        self.timestamps: list[float] = []

        # sanity checks
        if self.rate == 0:
            raise ValueError("Cooldown rate must be greater than 0")
        if self.interval == 0:
            raise ValueError("Cooldown interval must be greater than 0")

    def on_cooldown(self) -> bool:
        """
        Returns the cooldown state of the command.

        Returns:
            boolean state if the command is on cooldown or not

        """
        self._trim()

        return len(self.timestamps) >= self.rate

    def acquire_token(self) -> bool:
        """
        Attempt to acquire a token for a command to run.

        Returns:
            True if a token was acquired, False if not

        """
        self._trim()

        if len(self.timestamps) >= self.rate:
            return False

        self.timestamps.append(time.time())

        return True

    def get_cooldown_time(self) -> float:
        """
        Returns how long until the cooldown will reset.

        Returns:
            remaining cooldown time, will return 0 if the cooldown has not been reached

        """
        self._trim()

        if len(self.timestamps) < self.rate:
            return 0

        return self.timestamps[0] + self.interval - time.time()

    def reset(self) -> None:
        """Resets the timestamps for this cooldown."""
        self.timestamps = []

    def _trim(self) -> None:
        """Removes all timestamps that are outside the current interval."""
        cutoff = time.time() - self.interval

        while self.timestamps and self.timestamps[0] < cutoff:
            self.timestamps.pop(0)


class ExponentialBackoffSystem(CooldownSystem):
    """
    An exponential backoff cooldown strategy that doubles the interval between allowed commands after each failed attempt, up to a maximum interval.

    Attributes:
        rate: The number of commands allowed per interval.
        interval: The initial time window (in seconds) within which the allowed number of commands can be executed.
        max_interval: The maximum time window (in seconds) between allowed commands.
        multiplier: The multiplier to apply to the interval after each failed attempt.

    ??? tip "Example Use-case"
        This strategy is useful for scenarios where you want to progressively slow down repeated attempts at a command, such as preventing brute force attacks or limiting retries on failed operations.

    """

    def __init__(self, rate: int, interval: float, max_interval: float, multiplier: float = 2) -> None:
        super().__init__(rate, interval)
        self.max_interval = max_interval
        self.multiplier = multiplier

    def determine_cooldown(self) -> None:
        c_time = time.time()

        if c_time > self.opened + self.interval:
            if self.interval < self.max_interval:
                self.interval *= self.multiplier
            self.reset()


class LeakyBucketSystem(CooldownSystem):
    """
    A leaky bucket cooldown strategy that gradually replenishes tokens over time, allowing commands to be executed as long as there are available tokens in the bucket.

    Attributes:
        rate: The number of tokens generated per interval.
        interval: The time window (in seconds) within which the tokens are generated.

    ??? tip "Example Use-case"
        This strategy is useful for scenarios where you want to allow a steady flow of commands to be executed while preventing sudden bursts, such as rate limiting API calls or managing user interactions in a chatbot.

    """

    def determine_cooldown(self) -> None:
        c_time = time.time()

        tokens_to_recover = (c_time - self.opened) / self.interval
        if tokens_to_recover >= 1:
            self._tokens = min(self.rate, self._tokens + int(tokens_to_recover))
            self.opened = c_time


class TokenBucketSystem(CooldownSystem):
    """
    A token bucket cooldown strategy that generates tokens at a specific rate up to a burst rate, allowing commands to be executed as long as there are available tokens in the bucket.

    Attributes:
        rate: The number of tokens generated per interval.
        interval: The time window (in seconds) within which the tokens are generated.
        burst_rate: The maximum number of tokens that can be held in the bucket at any given time.

    ??? tip "Example Use-case"
        This strategy is useful for scenarios where you want to allow a burst of commands to be executed while limiting the overall rate, such as handling peak traffic in an API or permitting rapid user interactions in a game.

    """

    def __init__(self, rate: int, interval: float, burst_rate: int) -> None:
        super().__init__(rate, interval)
        self.burst_rate = burst_rate

    def determine_cooldown(self) -> None:
        c_time = time.time()

        tokens_to_recover = (c_time - self.opened) / self.interval
        if tokens_to_recover >= 1:
            self._tokens = min(self.burst_rate, self._tokens + int(tokens_to_recover))
            self.opened = c_time


class Cooldown:
    """
    Manages cooldowns and their respective buckets for a command.

    There are two pre-defined cooldown systems, a sliding window and a standard cooldown system (default);
    you can specify which one to use by passing in the cooldown_system parameter.

    Attributes:
        bucket: The bucket to use for this cooldown
        cooldown_repositories: A dictionary of cooldowns for each bucket
        rate: How many commands may be ran per interval
        interval: How many seconds to wait for a cooldown
        cooldown_system: The cooldown system to use for this cooldown

    """

    __slots__ = "bucket", "cooldown_repositories", "rate", "interval", "cooldown_system"

    def __init__(
        self,
        cooldown_bucket: Buckets,
        rate: int,
        interval: float,
        *,
        cooldown_system: Type[CooldownSystem] = CooldownSystem,
    ) -> None:
        self.bucket: Buckets = cooldown_bucket
        self.cooldown_repositories = {}
        self.rate: int = rate
        self.interval: float = interval

        self.cooldown_system: Type[CooldownSystem] = cooldown_system or CooldownSystem

    async def get_cooldown(self, context: "BaseContext") -> "CooldownSystem":
        key = await self.bucket(context)

        if key not in self.cooldown_repositories:
            cooldown = self.cooldown_system(self.rate, self.interval)
            self.cooldown_repositories[key] = cooldown
            return cooldown
        return self.cooldown_repositories.get(await self.bucket(context))

    def get_cooldown_with_key(self, key: Any, *, create: bool = False) -> typing.Optional["CooldownSystem"]:
        """
        Get the cooldown system for the command.

        Note:
            The preferred way to get the cooldown system is to use `get_cooldown` as it will use the context to get the correct key.

        Args:
            key: The key to get the cooldown system for
            create: Whether to create a new cooldown system if one does not exist

        """
        if key not in self.cooldown_repositories and create:
            cooldown = self.cooldown_system(self.rate, self.interval)
            self.cooldown_repositories[key] = cooldown
            return cooldown
        return self.cooldown_repositories.get(key)

    async def acquire_token(self, context: "BaseContext") -> bool:
        """
        Attempt to acquire a token for a command to run. Uses the context of the command to use the correct CooldownStrategy.

        Args:
            context: The context of the command

        Returns:
            True if a token was acquired, False if not

        """
        cooldown = await self.get_cooldown(context)

        return cooldown.acquire_token()

    async def get_cooldown_time(self, context: "BaseContext") -> float:
        """
        Get the remaining cooldown time.

        Args:
            context: The context of the command

        Returns:
            remaining cooldown time, will return 0 if the cooldown has not been reached

        """
        cooldown = await self.get_cooldown(context)
        return cooldown.get_cooldown_time()

    def get_cooldown_time_with_key(self, key: Any, *, create: bool = False) -> float:
        """
        Get the remaining cooldown time with a key instead of the context.

        Note:
            The preferred way to get the cooldown system is to use `get_cooldown` as it will use the context to get the correct key.

        Args:
            key: The key to get the cooldown system for
            create: Whether to create a new cooldown system if one does not exist

        """
        cooldown = self.get_cooldown_with_key(key, create=create)
        if cooldown is not None:
            return cooldown.get_cooldown_time()
        return 0

    async def on_cooldown(self, context: "BaseContext") -> bool:
        """
        Returns the cooldown state of the command.

        Args:
            context: The context of the command

        Returns:
            boolean state if the command is on cooldown or not

        """
        cooldown = await self.get_cooldown(context)
        return cooldown.on_cooldown()

    async def reset_all(self) -> None:
        """
        Resets this cooldown system to its initial state.

        !!! warning     To be clear, this will reset **all** cooldowns
        for this command to their initial states

        """
        # this doesnt need to be async, but for consistency, it is
        self.cooldown_repositories = {}

    async def reset(self, context: "BaseContext") -> None:
        """
        Resets the cooldown for the bucket of which invoked this command.

        Args:
            context: The context of the command

        """
        cooldown = await self.get_cooldown(context)
        cooldown.reset()

    def reset_with_key(self, key: Any) -> bool:
        """
        Resets the cooldown for the bucket associated with the provided key.

        Note:
            The preferred way to reset the cooldown system is to use `reset_cooldown` as it will use the context to reset the correct cooldown.

        Args:
            key: The key to reset the cooldown system for

        Returns:
            True if the key existed and was reset successfully, False if the key didn't exist.

        """
        cooldown = self.get_cooldown_with_key(key)
        if cooldown is not None:
            cooldown.reset()
            return True
        return False


class MaxConcurrency:
    """
    Limits how many instances of a command may be running concurrently.

    Attributes:
        bucket Buckets: The bucket this concurrency applies to
        concurrent int: The maximum number of concurrent instances permitted to
        wait bool: Should we wait until a instance is available

    """

    def __init__(self, concurrent: int, concurrency_bucket: Buckets, wait: bool = False) -> None:
        self.bucket: Buckets = concurrency_bucket
        self.concurrency_repository: Dict = {}
        self.concurrent: int = concurrent
        self.wait = wait

    async def get_semaphore(self, context: "BaseContext") -> asyncio.Semaphore:
        """
        Get the semaphore associated with the given context.

        Args:
            context: The commands context

        Returns:
            A semaphore object

        """
        key = await self.bucket(context)

        if key not in self.concurrency_repository:
            semaphore = asyncio.Semaphore(self.concurrent)
            self.concurrency_repository[key] = semaphore
            return semaphore
        return self.concurrency_repository.get(key)

    async def acquire(self, context: "BaseContext") -> bool:
        """
        Acquire an instance of the semaphore.

        Args:
            context:The context of the command
        Returns:
            If the semaphore was successfully acquired

        """
        semaphore = await self.get_semaphore(context)

        if not self.wait and semaphore.locked():
            return False
        return await semaphore.acquire()

    async def release(self, context: "BaseContext") -> None:
        """
        Release the semaphore.

        Args:
            context: The context of the command

        """
        semaphore = await self.get_semaphore(context)

        semaphore.release()



from typing import cast, Callable, Any

from interactions import Client, BaseContext, listen
from interactions.api.events import CallbackAdded, ExtensionUnload
from interactions.ext import prefixed_commands as prefixed

from .context import HybridContext
from .hybrid_slash import (
    _values_wrapper,
    base_subcommand_generator,
    HybridSlashCommand,
    _HybridToPrefixedCommand,
    slash_to_prefixed,
)

__all__ = ("HybridManager", "setup")


def add_use_slash_command_message(
    prefixed_cmd: _HybridToPrefixedCommand, slash_cmd: HybridSlashCommand
) -> _HybridToPrefixedCommand:
    if prefixed_cmd.has_binding:

        def wrap_old_callback(func: Callable) -> Any:
            async def _msg_callback(self, ctx: prefixed.PrefixedContext, *args, **kwargs):
                await ctx.reply(f"This command has been updated. Please use {slash_cmd.mention(ctx.guild_id)} instead.")
                await func(ctx, *args, **kwargs)

            return _msg_callback

    else:

        def wrap_old_callback(func: Callable) -> Any:
            async def _msg_callback(ctx: prefixed.PrefixedContext, *args, **kwargs):
                await ctx.reply(f"This command has been updated. Please use {slash_cmd.mention(ctx.guild_id)} instead.")
                await func(ctx, *args, **kwargs)

            return _msg_callback

    prefixed_cmd.callback = wrap_old_callback(prefixed_cmd.callback)
    return prefixed_cmd


class HybridManager:
    """
    The main part of the extension. Deals with injecting itself in the first place.

    Parameters:
        client: The client instance.
        hybrid_context: The object to instantiate for Hybrid Context
        use_slash_command_msg: If enabled, will send out a message encouraging users to use the slash command \
            equivalent whenever they use the prefixed command version.

    """

    def __init__(
        self, client: Client, *, hybrid_context: type[BaseContext] = HybridContext, use_slash_command_msg: bool = False
    ) -> None:
        if not hasattr(client, "prefixed") or not isinstance(client.prefixed, prefixed.PrefixedManager):
            raise TypeError("Prefixed commands are not set up for this bot.")

        self.hybrid_context = hybrid_context
        self.use_slash_command_msg = use_slash_command_msg

        self.client = cast(prefixed.PrefixedInjectedClient, client)
        self.ext_command_list: dict[str, list[str]] = {}

        self.client.add_listener(self.handle_ext_unload.copy_with_binding(self))

        self.client._add_command_hook.append(self._add_hybrid_command)

        self.client.hybrid = self

    @listen("on_callback_added")
    async def add_hybrid_command(self, event: CallbackAdded) -> None:
        # just here for backwards compatability since it was accidentially public, don't rely on it
        self._add_hybrid_command(event.callback)

    def _add_hybrid_command(self, callback: Callable):
        if not isinstance(callback, HybridSlashCommand) or not callback.callback or callback._dummy_base:
            return

        cmd = callback
        prefixed_transform = slash_to_prefixed(cmd)

        if self.use_slash_command_msg:
            prefixed_transform = add_use_slash_command_message(prefixed_transform, cmd)

        if cmd.is_subcommand:
            base = None
            if not (base := self.client.prefixed.commands.get(str(cmd.name))):
                base = base_subcommand_generator(
                    str(cmd.name),
                    list(_values_wrapper(cmd.name.to_locale_dict())) + cmd.aliases,
                    str(cmd.name),
                    group=False,
                )
                self.client.prefixed.add_command(base)

            if cmd.group_name:  # group command
                group = None
                if not (group := base.subcommands.get(str(cmd.group_name))):
                    group = base_subcommand_generator(
                        str(cmd.group_name),
                        list(_values_wrapper(cmd.group_name.to_locale_dict())) + cmd.aliases,
                        str(cmd.group_name),
                        group=True,
                    )
                    base.add_command(group)
                base = group

            # since this is added *after* the base command has been added to the bot, we need to run
            # this function ourselves
            prefixed_transform._parse_parameters()
            base.add_command(prefixed_transform)
        else:
            self.client.prefixed.add_command(prefixed_transform)

        if cmd.extension:
            self.ext_command_list.setdefault(cmd.extension.extension_name, []).append(cmd.resolved_name)

    @listen("extension_unload")
    async def handle_ext_unload(self, event: ExtensionUnload) -> None:
        if not self.ext_command_list.get(event.extension.extension_name):
            return

        for cmd in self.ext_command_list[event.extension.extension_name]:
            self.client.prefixed.remove_command(cmd, delete_parent_if_empty=True)

        del self.ext_command_list[event.extension.extension_name]


def setup(
    client: Client, *, hybrid_context: type[BaseContext] = HybridContext, use_slash_command_msg: bool = False
) -> HybridManager:
    """
    Sets up hybrid commands. It is recommended to use this function directly to do so.

    !!! warning
        Prefixed commands need to be set up prior to using this.

    Args:
        client: The client instance.
        hybrid_context: The object to instantiate for Hybrid Context
        use_slash_command_msg: If enabled, will send out a message encouraging users to use the slash command \
            equivalent whenever they use the prefixed command version.

    Returns:
        HybridManager: The class that deals with all things hybrid commands.

    """
    return HybridManager(client, hybrid_context=hybrid_context, use_slash_command_msg=use_slash_command_msg)



from functools import partial, total_ordering
from typing import Any, TYPE_CHECKING

import attrs
from interactions.client.const import MISSING, T, Missing
from interactions.client.utils import nulled_boolean_get
from interactions.client.utils.attr_converters import optional as optional_c
from interactions.client.utils.serializer import dict_filter, to_image_data
from interactions.models.discord.asset import Asset
from interactions.models.discord.file import UPLOADABLE_TYPE
from interactions.models.discord.color import COLOR_TYPES, Color, process_color
from interactions.models.discord.emoji import PartialEmoji
from interactions.models.discord.enums import Permissions
from .base import DiscordObject

if TYPE_CHECKING:
    from interactions.client import Client
    from interactions.models.discord.guild import Guild
    from interactions.models.discord.user import Member
    from interactions.models.discord.snowflake import Snowflake_Type

__all__ = ("Role",)


def sentinel_converter(value: bool | T | None, sentinel: T = attrs.NOTHING) -> bool | T:
    if value is sentinel:
        return False
    return True if value is None else value


@attrs.define(eq=False, order=False, hash=False, kw_only=True)
@total_ordering
class Role(DiscordObject):
    _sentinel = object()

    name: str = attrs.field(repr=True)
    color: "Color" = attrs.field(repr=False, converter=Color)
    hoist: bool = attrs.field(repr=False, default=False)
    position: int = attrs.field(repr=True)
    permissions: "Permissions" = attrs.field(repr=False, converter=Permissions)
    managed: bool = attrs.field(repr=False, default=False)
    mentionable: bool = attrs.field(repr=False, default=True)
    premium_subscriber: bool = attrs.field(
        repr=False, default=_sentinel, converter=partial(sentinel_converter, sentinel=_sentinel)
    )
    subscription_listing_id: "Snowflake_Type | None" = attrs.field(default=None, repr=False)
    purchasable_or_has_subscribers: bool = attrs.field(default=False)
    _icon: Asset | None = attrs.field(repr=False, default=None)
    _unicode_emoji: PartialEmoji | None = attrs.field(
        repr=False, default=None, converter=optional_c(PartialEmoji.from_str)
    )
    _guild_id: "Snowflake_Type" = attrs.field(
        repr=False,
    )
    _bot_id: "Snowflake_Type | None" = attrs.field(repr=False, default=None)
    _integration_id: "Snowflake_Type | None" = attrs.field(repr=False, default=None)  # todo integration object?
    _guild_connections: bool = attrs.field(repr=False, default=False)

    def __lt__(self: "Role", other: "Role") -> bool:
        if not isinstance(self, Role) or not isinstance(other, Role):
            return NotImplemented

        if self._guild_id != other._guild_id:
            raise RuntimeError("Unable to compare Roles from different guilds.")

        if self.id == self._guild_id:  # everyone role
            # everyone role is on the bottom, so check if the other role is, well, not it
            # because then it must be higher than it
            return other.id != self.id

        if self.position < other.position:
            return True

        return self.id < other.id if self.position == other.position else False

    @classmethod
    def _process_dict(cls, data: dict[str, Any], client: "Client") -> dict[str, Any]:
        data |= data.pop("tags", {})

        if icon_hash := data.get("icon"):
            data["icon"] = Asset.from_path_hash(client, f"role-icons/{data['id']}/{{}}", icon_hash)

        data["premium_subscriber"] = nulled_boolean_get(data, "premium_subscriber")
        data["guild_connections"] = nulled_boolean_get(data, "guild_connections")
        data["available_for_purchase"] = nulled_boolean_get(data, "available_for_purchase")

        return data

    async def fetch_bot(self, *, force: bool = False) -> "Member | None":
        """
        Fetch the bot associated with this role if any.

        Args:
            force: Whether to force fetch the bot from the API.

        Returns:
            Member object if any

        """
        if self._bot_id is None:
            return None
        return await self._client.cache.fetch_member(self._guild_id, self._bot_id, force=force)

    def get_bot(self) -> "Member | None":
        """
        Get the bot associated with this role if any.

        Returns:
            Member object if any

        """
        if self._bot_id is None:
            return None
        return self._client.cache.get_member(self._guild_id, self._bot_id)

    @property
    def guild(self) -> "Guild":
        """The guild object this role is from."""
        return self._client.cache.get_guild(self._guild_id)  # pyright: ignore [reportGeneralTypeIssues]

    @property
    def default(self) -> bool:
        """Is this the `@everyone` role."""
        return self.id == self._guild_id

    @property
    def bot_managed(self) -> bool:
        """Is this role owned/managed by a bot."""
        return self._bot_id is not None

    @property
    def is_linked_role(self) -> bool:
        """Is this role a linked role."""
        return self._guild_connections

    @property
    def mention(self) -> str:
        """Returns a string that would mention the role."""
        return f"<@&{self.id}>" if self.id != self._guild_id else "@everyone"

    @property
    def integration(self) -> bool:
        """Is this role owned/managed by an integration."""
        return self._integration_id is not None

    @property
    def members(self) -> list["Member"]:
        """List of members with this role"""
        return [member for member in self.guild.members if member.has_role(self)]

    @property
    def icon(self) -> Asset | PartialEmoji | None:
        """
        The icon of this role

        !!! note
            You have to use this method instead of the `_icon` attribute, because the first does account for unicode emojis
        """
        return self._icon or self._unicode_emoji

    @property
    def is_assignable(self) -> bool:
        """
        Can this role be assigned or removed by this bot?

        !!! note
            This does not account for permissions, only the role hierarchy

        """
        return (self.default or self.guild.me.top_role > self) and not self.managed

    async def delete(self, reason: str | Missing = MISSING) -> None:
        """
        Delete this role.

        Args:
            reason: An optional reason for this deletion

        """
        await self._client.http.delete_guild_role(self._guild_id, self.id, reason)

    async def edit(
        self,
        *,
        name: str | None = None,
        permissions: str | None = None,
        color: Color | COLOR_TYPES | None = None,
        hoist: bool | None = None,
        mentionable: bool | None = None,
        icon: bytes | UPLOADABLE_TYPE | None = None,
        unicode_emoji: str | None = None,
    ) -> "Role":
        """
        Edit this role, all arguments are optional.

        Args:
            name: name of the role
            permissions: New permissions to use
            color: The color of the role
            hoist: whether the role should be displayed separately in the sidebar
            mentionable: whether the role should be mentionable
            icon: (Guild Level 2+) Bytes-like object representing the icon; supports PNG, JPEG and WebP
            unicode_emoji: (Guild Level 2+) Unicode emoji for the role; can't be used with icon

        Returns:
            Role with updated information

        """
        color = process_color(color)

        if icon and unicode_emoji:
            raise ValueError("Cannot pass both icon and unicode_emoji")
        if icon:
            icon = to_image_data(icon)

        payload = dict_filter(
            {
                "name": name,
                "permissions": permissions,
                "color": color,
                "hoist": hoist,
                "mentionable": mentionable,
                "icon": icon,
                "unicode_emoji": unicode_emoji,
            }
        )

        r_data = await self._client.http.modify_guild_role(self._guild_id, self.id, payload)
        r_data = dict(r_data)  # to convert typed dict to regular dict
        r_data["guild_id"] = self._guild_id
        return self.from_dict(r_data, self._client)

    async def move(self, position: int, reason: str | Missing = MISSING) -> "Role":
        """
        Move this role to a new position.

        Args:
            position: The new position of the role
            reason: An optional reason for this move

        Returns:
            The role object

        """
        await self._client.http.modify_guild_role_positions(
            self._guild_id, [{"id": self.id, "position": position}], reason
        )
        return self



import asyncio
from abc import ABC, abstractmethod
from asyncio import QueueEmpty
from collections.abc import AsyncIterator as _AsyncIterator
from typing import List, Any

from interactions.client.const import MISSING, Absent
from interactions.models.discord import snowflake

__all__ = ("AsyncIterator",)


class AsyncIterator(_AsyncIterator, ABC):
    def __init__(self, limit: int = 50) -> None:
        self._queue: asyncio.Queue = asyncio.Queue()
        """The queue of items in the iterator"""

        self._limit: int = limit or MISSING
        """the limit of items to retrieve"""

        self.last: Absent[Any] = MISSING
        """The last item retrieved"""

        self._retrieved_objects: List = []
        """All items this iterator has retrieved"""

    @property
    def _continue(self) -> bool:
        """Whether iteration should continue. Returns False if the limit has been reached."""
        return len(self._retrieved_objects) < self._limit if self._limit else True

    @property
    def get_limit(self) -> int:
        """Get how the maximum number of items that should be retrieved."""
        return min(self._limit - len(self._retrieved_objects), 100) if self._limit else 100

    @property
    def total_retrieved(self) -> int:
        """Get the total number of objects this iterator has retrieved."""
        return len(self._retrieved_objects)

    async def add_object(self, obj) -> None:
        """Add an object to iterator's queue."""
        return await self._queue.put(obj)

    @abstractmethod
    async def fetch(self) -> list:
        """
        Fetch additional objects.

        Your implementation of this method *must* return a list of objects.
        If no more objects are available, raise QueueEmpty

        Returns:
            List of objects

        Raises:
            QueueEmpty:  when no more objects are available.

        """
        ...

    async def _get_items(self) -> None:
        if self._continue:
            data = await self.fetch()
            [await self.add_object(obj) for obj in data]
        else:
            raise QueueEmpty

    async def __anext__(self) -> Any:
        try:
            if self._queue.empty():
                await self._get_items()
            self.last = self._queue.get_nowait()

            # add the message to the already retrieved objects, so that the search function works when calling it multiple times
            self._retrieved_objects.append(self.last)

            return self.last
        except QueueEmpty as e:
            raise StopAsyncIteration from e

    async def flatten(self) -> List:
        """Flatten this iterator into a list of objects."""
        return [elem async for elem in self]

    async def search(self, target_id: "snowflake.Snowflake_Type") -> bool:
        """Search the iterator for an object with the given ID."""
        target_id = snowflake.to_snowflake(target_id)

        if target_id in [o.id for o in self._retrieved_objects]:
            return True

        async for o in self:
            if o.id == target_id:
                return True
        return False



import copy
from typing import TYPE_CHECKING

import interactions.api.events as events
from interactions.client.const import MISSING
from interactions.models import ScheduledEvent
from ._template import EventMixinTemplate, Processor

if TYPE_CHECKING:
    from interactions.api.events import RawGatewayEvent

__all__ = ("ScheduledEvents",)


class ScheduledEvents(EventMixinTemplate):
    @Processor.define()
    async def _on_raw_guild_scheduled_event_create(self, event: "RawGatewayEvent") -> None:
        scheduled_event = self.cache.place_scheduled_event_data(event.data)

        self.dispatch(events.GuildScheduledEventCreate(scheduled_event))

    @Processor.define()
    async def _on_raw_guild_scheduled_event_update(self, event: "RawGatewayEvent") -> None:
        before = copy.copy(self.cache.get_scheduled_event(event.data.get("id")))
        after = self.cache.place_scheduled_event_data(event.data)

        self.dispatch(events.GuildScheduledEventUpdate(before or MISSING, after))

    @Processor.define()
    async def _on_raw_guild_scheduled_event_delete(self, event: "RawGatewayEvent") -> None:
        # for some reason this event returns the deleted scheduled event data?
        # so we create an object from it
        scheduled_event = ScheduledEvent.from_dict(event.data, self)
        self.cache.delete_scheduled_event(event.data.get("id"))

        self.dispatch(events.GuildScheduledEventDelete(scheduled_event))

    @Processor.define()
    async def _on_raw_guild_scheduled_event_user_add(self, event: "RawGatewayEvent") -> None:
        self.dispatch(
            events.GuildScheduledEventUserAdd(
                event.data["guild_id"], event.data["guild_scheduled_event_id"], event.data["user_id"]
            )
        )

    @Processor.define()
    async def _on_raw_guild_scheduled_event_user_remove(self, event: "RawGatewayEvent") -> None:
        self.dispatch(
            events.GuildScheduledEventUserRemove(
                event.data["guild_id"], event.data["guild_scheduled_event_id"], event.data["user_id"]
            )
        )



import asyncio
from io import BytesIO
from typing import Optional, TYPE_CHECKING

import attrs
from discord_typings import VoiceStateData

from interactions.api.voice.player import Player
from interactions.api.voice.recorder import Recorder
from interactions.api.voice.voice_gateway import VoiceGateway
from interactions.client.const import MISSING, Missing
from interactions.client.errors import VoiceAlreadyConnected, VoiceConnectionTimeout
from interactions.client.utils import optional
from interactions.models.discord.enums import Intents
from interactions.models.discord.snowflake import Snowflake_Type, to_snowflake
from interactions.models.discord.voice_state import VoiceState

if TYPE_CHECKING:
    from interactions.api.voice.audio import BaseAudio
    from interactions.api.gateway.gateway import GatewayClient

__all__ = ("ActiveVoiceState",)


@attrs.define(eq=False, order=False, hash=False, kw_only=True)
class ActiveVoiceState(VoiceState):
    ws: Optional[VoiceGateway] = attrs.field(repr=False, default=None)
    """The websocket for this voice state"""
    player: Optional[Player] = attrs.field(repr=False, default=None)
    """The playback task that broadcasts audio data to discord"""
    recorder: Optional[Recorder] = attrs.field(default=None)
    """A recorder task to capture audio from discord"""
    _volume: float = attrs.field(repr=False, default=0.5)

    # standard voice states expect this data, this voice state lacks it initially; so we make them optional
    user_id: "Snowflake_Type" = attrs.field(repr=False, default=MISSING, converter=optional(to_snowflake))
    _guild_id: Optional["Snowflake_Type"] = attrs.field(repr=False, default=None, converter=optional(to_snowflake))
    _member_id: Optional["Snowflake_Type"] = attrs.field(repr=False, default=None, converter=optional(to_snowflake))

    def __attrs_post_init__(self) -> None:
        # jank line to handle the two inherently incompatible data structures
        self._member_id = self.user_id = self._client.user.id

    def __del__(self) -> None:
        if self.connected:
            self.ws.close()
        if self.player:
            self.player.stop()

    def __repr__(self) -> str:
        return f"<ActiveVoiceState: channel={self.channel} guild={self.guild} volume={self.volume} playing={self.playing} audio={self.current_audio}>"

    @property
    def current_audio(self) -> Optional["BaseAudio"]:
        """The current audio being played"""
        if self.player:
            return self.player.current_audio

    @property
    def volume(self) -> float:
        """Get the volume of the player"""
        return self._volume

    @volume.setter
    def volume(self, value) -> None:
        """Set the volume of the player"""
        if value < 0.0:
            raise ValueError("Volume may not be negative.")
        self._volume = value
        if self.player and hasattr(self.player.current_audio, "volume"):
            self.player.current_audio.volume = value

    @property
    def paused(self) -> bool:
        """Is the player currently paused"""
        return self.player.paused if self.player else False

    @property
    def playing(self) -> bool:
        """Are we currently playing something?"""
        # noinspection PyProtectedMember
        return bool(self.player and self.current_audio and not self.player.stopped and self.player._resume.is_set())

    @property
    def stopped(self) -> bool:
        """Is the player stopped?"""
        return self.player.stopped if self.player else True

    @property
    def connected(self) -> bool:
        """Is this voice state currently connected?"""
        # noinspection PyProtectedMember
        return False if self.ws is None else self.ws._closed.is_set()

    @property
    def gateway(self) -> "GatewayClient":
        return self._client.get_guild_websocket(self._guild_id)

    async def wait_for_stopped(self) -> None:
        """Wait for the player to stop playing."""
        if self.player:
            # noinspection PyProtectedMember
            await self.player._stopped.wait()

    async def _ws_connect(self) -> None:
        """Runs the voice gateway connection"""
        async with self.ws:
            try:
                await self.ws.run()
            finally:
                if self.playing:
                    await self.stop()

    async def ws_connect(self) -> None:
        """Connect to the voice gateway for this voice state"""
        self.ws = VoiceGateway(self._client._connection_state, self._voice_state.data, self._voice_server.data)

        _ = asyncio.create_task(self._ws_connect())  # noqa: RUF006
        await self.ws.wait_until_ready()

    def _guild_predicate(self, event) -> bool:
        return int(event.data["guild_id"]) == self._guild_id

    async def connect(self, timeout: int = 5) -> None:
        """
        Establish the voice connection.

        Args:
            timeout: How long to wait for state and server information from discord

        Raises:
            VoiceAlreadyConnected: if the voice state is already connected to the voice channel
            VoiceConnectionTimeout: if the voice state fails to connect

        """
        if self.connected:
            raise VoiceAlreadyConnected

        if Intents.GUILD_VOICE_STATES not in self._client.intents:
            raise RuntimeError("Cannot connect to voice without the GUILD_VOICE_STATES intent.")

        tasks = [
            asyncio.create_task(
                self._client.wait_for("raw_voice_state_update", self._guild_predicate, timeout=timeout)
            ),
            asyncio.create_task(
                self._client.wait_for("raw_voice_server_update", self._guild_predicate, timeout=timeout)
            ),
        ]

        await self.gateway.voice_state_update(self._guild_id, self._channel_id, self.self_mute, self.self_deaf)

        self.logger.debug("Waiting for voice connection data...")

        try:
            self._voice_state, self._voice_server = await asyncio.gather(*tasks)
        except asyncio.TimeoutError:
            raise VoiceConnectionTimeout from None

        self.logger.debug("Attempting to initialise voice gateway...")
        await self.ws_connect()

    async def disconnect(self) -> None:
        """Disconnect from the voice channel."""
        await self.gateway.voice_state_update(self._guild_id, None)

    async def move(self, channel: "Snowflake_Type", timeout: int = 5) -> None:
        """
        Move to another voice channel.

        Args:
            channel: The channel to move to
            timeout: How long to wait for state and server information from discord

        """
        target_channel = to_snowflake(channel)
        if target_channel != self._channel_id:
            already_paused = self.paused
            if self.player:
                self.player.pause()

            self._channel_id = target_channel
            await self.gateway.voice_state_update(self._guild_id, self._channel_id, self.self_mute, self.self_deaf)

            self.logger.debug("Waiting for voice connection data...")
            try:
                await self._client.wait_for("raw_voice_state_update", self._guild_predicate, timeout=timeout)
            except asyncio.TimeoutError:
                await self._close_connection()
                raise VoiceConnectionTimeout from None

            if self.player and not already_paused:
                self.player.resume()

    async def stop(self) -> None:
        """Stop playback."""
        self.player.stop()
        await self.player._stopped.wait()

    def pause(self) -> None:
        """Pause playback"""
        self.player.pause()

    def resume(self) -> None:
        """Resume playback."""
        self.player.resume()

    async def play(self, audio: "BaseAudio") -> None:
        """
        Start playing an audio object.

        Waits for the player to stop before returning.

        Args:
            audio: The audio object to play

        """
        if self.player:
            await self.stop()

        with Player(audio, self, asyncio.get_running_loop()) as self.player:
            self.player.play()
            await self.wait_for_stopped()

    def play_no_wait(self, audio: "BaseAudio") -> asyncio.Task:
        """
        Start playing an audio object, but don't wait for playback to finish.

        Args:
            audio: The audio object to play

        """
        return asyncio.create_task(self.play(audio))

    def create_recorder(self) -> Recorder:
        """Create a recorder instance."""
        if not self.recorder:
            self.recorder = Recorder(self, asyncio.get_running_loop())
        return self.recorder

    async def start_recording(self, encoding: Optional[str] = None, *, output_dir: str | Missing = Missing) -> Recorder:
        """
        Start recording the voice channel.

        If no recorder exists, one will be created.

        Args:
            encoding: What format the audio should be encoded to.
            output_dir: The directory to save the audio to

        """
        if not self.recorder:
            self.recorder = Recorder(self, asyncio.get_running_loop())

        if self.recorder.used:
            if self.recorder.recording:
                raise RuntimeError("Another recording is still in progress, please stop it first.")
            self.recorder = Recorder(self, asyncio.get_running_loop())

        if encoding is not None:
            self.recorder.encoding = encoding

        await self.recorder.start_recording(output_dir=output_dir)
        return self.recorder

    async def stop_recording(self) -> dict[int, BytesIO]:
        """
        Stop the recording.

        Returns:
            dict[snowflake, BytesIO]: The recorded audio

        """
        if not self.recorder or not self.recorder.recording or not self.recorder.audio:
            raise RuntimeError("No recorder is running!")
        await self.recorder.stop_recording()

        self.recorder.audio.finished.wait()
        return self.recordings

    @property
    def recordings(self) -> dict[int, BytesIO]:
        return self.recorder.output if self.recorder else {}

    async def _voice_server_update(self, data) -> None:
        """
        An internal receiver for voice server events.

        Args:
            data: voice server data

        """
        self.ws.set_new_voice_server(data)

    async def _voice_state_update(
        self,
        before: Optional[VoiceState],
        after: Optional[VoiceState],
        data: Optional[VoiceStateData],
    ) -> None:
        """
        An internal receiver for voice server state events.

        Args:
            before: The previous voice state
            after: The current voice state
            data: Raw data from gateway

        """
        if after is None:
            # bot disconnected
            self.logger.info(f"Disconnecting from voice channel {self._channel_id}")
            await self._close_connection()
            self._client.cache.delete_bot_voice_state(self._guild_id)
            return

        self.update_from_dict(data)

    async def _close_connection(self) -> None:
        """Close the voice connection."""
        if self.playing:
            await self.stop()
        if self.connected:
            self.ws.close()



from typing import TYPE_CHECKING, Any, List

import discord_typings

from interactions.client.const import MISSING, Absent
from interactions.models.discord.snowflake import to_snowflake
from ..route import Route

__all__ = ("ReactionRequests",)


if TYPE_CHECKING:
    from interactions.models.discord.snowflake import Snowflake_Type


class ReactionRequests:
    request: Any

    async def create_reaction(self, channel_id: "Snowflake_Type", message_id: "Snowflake_Type", emoji: str) -> None:
        """
        Create a reaction for a message.

        Args:
            channel_id: The channel this is taking place in
            message_id: The message to create a a reaction on
            emoji: The emoji to use (format: `name:id`)

        """
        return await self.request(
            Route(
                "PUT",
                "/channels/{channel_id}/messages/{message_id}/reactions/{emoji}/@me",
                channel_id=channel_id,
                message_id=message_id,
                emoji=emoji,
            )
        )

    async def remove_self_reaction(
        self, channel_id: "Snowflake_Type", message_id: "Snowflake_Type", emoji: str
    ) -> None:
        """
        Remove client's reaction from a message.

        Args:
            channel_id: The channel this is taking place in.
            message_id: The message to remove the reaction on.
            emoji: The emoji to remove. (format: `name:id`)

        """
        return await self.request(
            Route(
                "DELETE",
                "/channels/{channel_id}/messages/{message_id}/reactions/{emoji}/@me",
                channel_id=channel_id,
                message_id=message_id,
                emoji=emoji,
            )
        )

    async def remove_user_reaction(
        self,
        channel_id: "Snowflake_Type",
        message_id: "Snowflake_Type",
        emoji: str,
        user_id: "Snowflake_Type",
    ) -> None:
        """
        Remove user's reaction from a message.

        Args:
            channel_id: The channel this is taking place in
            message_id: The message to remove the reaction on.
            emoji: The emoji to remove. (format: `name:id`)
            user_id: The user to remove reaction of.

        """
        return await self.request(
            Route(
                "DELETE",
                "/channels/{channel_id}/messages/{message_id}/reactions/{emoji}/{user_id}",
                channel_id=channel_id,
                message_id=message_id,
                emoji=emoji,
                user_id=user_id,
            )
        )

    async def clear_reaction(self, channel_id: "Snowflake_Type", message_id: "Snowflake_Type", emoji: str) -> None:
        """
        Remove specific reaction from a message.

        Args:
            channel_id: The channel this is taking place in.
            message_id: The message to remove the reaction on.
            emoji: The emoji to remove. (format: `name:id`)

        """
        return await self.request(
            Route(
                "DELETE",
                "/channels/{channel_id}/messages/{message_id}/reactions/{emoji}",
                channel_id=channel_id,
                message_id=message_id,
                emoji=emoji,
            )
        )

    async def clear_reactions(self, channel_id: "Snowflake_Type", message_id: "Snowflake_Type") -> None:
        """
        Remove reactions from a message.

        Args:
            channel_id: The channel this is taking place in.
            message_id: The message to clear reactions from.

        """
        return await self.request(
            Route(
                "DELETE",
                "/channels/{channel_id}/messages/{message_id}/reactions",
                channel_id=channel_id,
                message_id=message_id,
            )
        )

    async def get_reactions(
        self,
        channel_id: "Snowflake_Type",
        message_id: "Snowflake_Type",
        emoji: str,
        limit: Absent[int] = MISSING,
        after: "Snowflake_Type" = MISSING,
    ) -> List[discord_typings.UserData]:
        """
        Gets specific reaction from a message.

        Args:
            channel_id: The channel this is taking place in.
            message_id: The message to get the reaction.
            emoji: The emoji to get. (format: `name:id`)
            limit: max number of entries to get
            after: snowflake to get entries after

        Returns:
            List of users who reacted with the emoji.

        """
        return await self.request(
            Route(
                "GET",
                "/channels/{channel_id}/messages/{message_id}/reactions/{emoji}",
                channel_id=to_snowflake(channel_id),
                message_id=to_snowflake(message_id),
                emoji=emoji,
            ),
            params={"limit": limit, "after": after},
        )



from typing import Any, Dict, List, Optional, Union

import attrs

from interactions.client.const import MISSING, Absent
from interactions.client.mixins.serialization import DictSerializationMixin
from interactions.client.utils.attr_converters import optional
from interactions.models.discord.base import ClientObject
from interactions.models.discord.emoji import PartialEmoji, process_emoji
from interactions.models.discord.enums import OnboardingMode, OnboardingPromptType
from interactions.models.discord.snowflake import (
    Snowflake,
    Snowflake_Type,
    SnowflakeObject,
    to_snowflake,
    to_snowflake_list,
)

__all__ = ("OnboardingPromptOption", "OnboardingPrompt", "Onboarding")


@attrs.define(eq=False, order=False, hash=False, kw_only=True)
class OnboardingPromptOption(SnowflakeObject, DictSerializationMixin):
    channel_ids: List["Snowflake"] = attrs.field(repr=False, converter=to_snowflake_list)
    """IDs for channels a member is added to when the option is selected"""
    role_ids: List["Snowflake"] = attrs.field(repr=False, converter=to_snowflake_list)
    """IDs for roles assigned to a member when the option is selected"""
    title: str = attrs.field(repr=False)
    """Title of the option"""
    description: Optional[str] = attrs.field(repr=False, default=None)
    """Description of the option"""
    emoji: Optional[PartialEmoji] = attrs.field(repr=False, default=None, converter=optional(PartialEmoji.from_dict))
    """Emoji of the option"""

    # this method is here because Discord needs the id field to be present in the payload
    @classmethod
    def create(
        cls,
        title: str,
        *,
        channel_ids: Optional[List[Snowflake_Type]] = None,
        role_ids: Optional[List[Snowflake_Type]] = None,
        description: Optional[str] = None,
        emoji: Optional[Union[PartialEmoji, dict, str]] = None,
    ) -> "OnboardingPromptOption":
        """
        Creates a new Onboarding prompt option object.

        Args:
            title: Title of the option
            channel_ids: Channel IDs that this option represents
            role_ids: Role IDs that this option represents
            description: Description of the option
            emoji: Emoji of the option

        Returns:
            The newly created OnboardingPromptOption object

        """
        return cls(
            id=0,
            channel_ids=channel_ids or [],
            role_ids=role_ids or [],
            title=title,
            description=description,
            emoji=process_emoji(emoji),
        )

    def as_dict(self) -> Dict[str, Any]:
        data = {
            "id": self.id,
            "channel_ids": self.channel_ids,
            "role_ids": self.role_ids,
            "title": self.title,
            "description": self.description,
        }
        # use the separate fields when sending to Discord
        if self.emoji is not None:
            data["emoji_id"] = self.emoji.id
            data["emoji_name"] = self.emoji.name
            data["emoji_animated"] = self.emoji.animated
        return data


@attrs.define(eq=False, order=False, hash=False, kw_only=False)
class OnboardingPrompt(SnowflakeObject, DictSerializationMixin):
    type: OnboardingPromptType = attrs.field(repr=False, converter=OnboardingPromptType)
    """Type of the prompt"""
    options: List[OnboardingPromptOption] = attrs.field(repr=False, converter=OnboardingPromptOption.from_list)
    """Options available in the prompt"""
    title: str = attrs.field(repr=False)
    """Title of the prompt"""
    single_select: bool = attrs.field(repr=False)
    """Whether users are limited to selecting one option for the prompt"""
    required: bool = attrs.field(repr=False)
    """Whether users are required to complete this prompt"""
    in_onboarding: bool = attrs.field(repr=False)
    """Whether the prompt is present in the onboarding flow; otherwise it is only in the Channels & Roles tab"""

    @classmethod
    def create(
        cls,
        *,
        type: Union[OnboardingPromptType, int] = OnboardingPromptType.MULTIPLE_CHOICE,
        options: List[OnboardingPromptOption],
        title: str,
        single_select: bool = False,
        required: bool = False,
        in_onboarding: bool = True,
    ) -> "OnboardingPrompt":
        """
        Creates a new Onboarding prompt object.

        Args:
            type: Type of the prompt
            options: Options available in the prompt
            title: Title of the prompt
            single_select: Whether users are limited to selecting one option for the prompt
            required: Whether users are required to complete this prompt
            in_onboarding: Whether the prompt is present in the onboarding flow; otherwise it is only in the Channels & Roles tab

        Returns:
            The newly created OnboardingPrompt object

        """
        return cls(
            id=0,
            type=type,
            options=options,
            title=title,
            single_select=single_select,
            required=required,
            in_onboarding=in_onboarding,
        )


@attrs.define(eq=False, order=False, hash=False, kw_only=False)
class Onboarding(ClientObject):
    """Represents the onboarding flow for a guild."""

    guild_id: Snowflake = attrs.field(repr=False, converter=to_snowflake)
    """ID of the guild this onboarding is part of"""
    prompts: list[OnboardingPrompt] = attrs.field(repr=False, converter=OnboardingPrompt.from_list)
    """Prompts shown during onboarding and in customize community"""
    default_channel_ids: list[Snowflake] = attrs.field(repr=False, converter=to_snowflake_list)
    """Channel IDs that members get opted into automatically"""
    enabled: bool = attrs.field(repr=False)
    """Whether onboarding is enabled in the guild"""
    mode: OnboardingMode = attrs.field(repr=False, converter=OnboardingMode)
    """Current mode of onboarding"""

    async def edit(
        self,
        *,
        prompts: Absent[List[OnboardingPrompt]] = MISSING,
        default_channel_ids: Absent[list[Snowflake_Type]] = MISSING,
        enabled: Absent[bool] = MISSING,
        mode: Absent[Union[OnboardingMode, int]] = MISSING,
        reason: Absent[str] = MISSING,
    ) -> None:
        """
        Edits this Onboarding flow.

        Args:
            prompts: Prompts shown during onboarding and in customize community
            default_channel_ids: Channel IDs that members get opted into automatically
            enabled: Whether onboarding is enabled in the guild
            mode: Current mode of onboarding
            reason: The reason for this change

        """
        payload = {
            "prompts": prompts,
            "default_channel_ids": default_channel_ids,
            "enabled": enabled,
            "mode": mode,
        }
        data = await self._client.http.modify_guild_onboarding(self.guild_id, payload, reason)
        self.update_from_dict(data)



from typing import Union, List, Optional, Type, TYPE_CHECKING

import interactions.models as models
from interactions.models.internal.application_commands import SlashCommandOption

__all__ = (
    "slash_attachment_option",
    "slash_bool_option",
    "slash_channel_option",
    "slash_float_option",
    "slash_int_option",
    "slash_mentionable_option",
    "slash_role_option",
    "slash_str_option",
    "slash_user_option",
)


if TYPE_CHECKING:
    from interactions.models.internal import SlashCommandChoice
    from interactions.models.discord import (
        User,
        Member,
        Role,
        BaseChannel,
        ChannelType,
        Attachment,
    )


def slash_str_option(
    description: str,
    required: bool = False,
    autocomplete: bool = False,
    choices: List[Union["SlashCommandChoice", dict]] | None = None,
    min_length: Optional[int] = None,
    max_length: Optional[int] = None,
    name: Optional[str] = None,
) -> Type[str]:
    """
    Annotates an argument as a string type slash command option.

    Args:
        description: The description of your option
        required: Is this option required?
        autocomplete: Use autocomplete for this option
        choices: The choices allowed by this command
        min_length: The minimum length of text a user can input.
        max_length: The maximum length of text a user can input.
        name: The name of the option. Defaults to the name of the argument

    """
    return SlashCommandOption(
        name=name,
        description=description,
        required=required,
        autocomplete=autocomplete,
        choices=choices or [],
        max_length=max_length,
        min_length=min_length,
        type=models.OptionType.STRING,
    )


def slash_float_option(
    description: str,
    required: bool = False,
    autocomplete: bool = False,
    choices: List[Union["SlashCommandChoice", dict]] | None = None,
    min_value: Optional[float] = None,
    max_value: Optional[float] = None,
    name: Optional[str] = None,
) -> Type[float]:
    """
    Annotates an argument as a float type slash command option.

    Args:
        description: The description of your option
        required: Is this option required?
        autocomplete: Use autocomplete for this option
        choices: The choices allowed by this command
        min_value: The minimum number allowed
        max_value: The maximum number allowed
        name: The name of the option. Defaults to the name of the argument

    """
    return SlashCommandOption(
        name=name,
        description=description,
        required=required,
        autocomplete=autocomplete,
        choices=choices or [],
        max_value=max_value,
        min_value=min_value,
        type=models.OptionType.NUMBER,
    )


def slash_int_option(
    description: str,
    required: bool = False,
    autocomplete: bool = False,
    choices: List[Union["SlashCommandChoice", dict]] | None = None,
    min_value: Optional[float] = None,
    max_value: Optional[float] = None,
    name: Optional[str] = None,
) -> Type[int]:
    """
    Annotates an argument as a integer type slash command option.

    Args:
        description: The description of your option
        required: Is this option required?
        autocomplete: Use autocomplete for this option
        choices: The choices allowed by this command
        min_value: The minimum number allowed
        max_value: The maximum number allowed
        name: The name of the option. Defaults to the name of the argument

    """
    return SlashCommandOption(
        name=name,
        description=description,
        required=required,
        autocomplete=autocomplete,
        choices=choices or [],
        max_value=max_value,
        min_value=min_value,
        type=models.OptionType.INTEGER,
    )


def slash_bool_option(
    description: str,
    required: bool = False,
    name: Optional[str] = None,
) -> Type[bool]:
    """
    Annotates an argument as a boolean type slash command option.

    Args:
        description: The description of your option
        required: Is this option required?
        name: The name of the option. Defaults to the name of the argument

    """
    return SlashCommandOption(
        name=name,
        description=description,
        required=required,
        type=models.OptionType.BOOLEAN,
    )


def slash_user_option(
    description: str,
    required: bool = False,
    autocomplete: bool = False,
    name: Optional[str] = None,
) -> Type[Union["User", "Member"]]:
    """
    Annotates an argument as a user type slash command option.

    Args:
        description: The description of your option
        required: Is this option required?
        autocomplete: Use autocomplete for this option
        name: The name of the option. Defaults to the name of the argument

    """
    return SlashCommandOption(
        name=name,
        description=description,
        required=required,
        autocomplete=autocomplete,
        type=models.OptionType.USER,
    )


def slash_channel_option(
    description: str,
    required: bool = False,
    autocomplete: bool = False,
    choices: List[Union["SlashCommandChoice", dict]] | None = None,
    channel_types: Optional[list[Union["ChannelType", int]]] = None,
    name: Optional[str] = None,
) -> Type["BaseChannel"]:
    """
    Annotates an argument as a channel type slash command option.

    Args:
        description: The description of your option
        required: Is this option required?
        autocomplete: Use autocomplete for this option
        choices: The choices allowed by this command
        channel_types: The types of channel allowed by this option
        name: The name of the option. Defaults to the name of the argument

    """
    return SlashCommandOption(
        name=name,
        description=description,
        required=required,
        autocomplete=autocomplete,
        choices=choices or [],
        channel_types=channel_types,
        type=models.OptionType.CHANNEL,
    )


def slash_role_option(
    description: str,
    required: bool = False,
    autocomplete: bool = False,
    choices: List[Union["SlashCommandChoice", dict]] | None = None,
    name: Optional[str] = None,
) -> Type["Role"]:
    """
    Annotates an argument as a role type slash command option.

    Args:
        description: The description of your option
        required: Is this option required?
        autocomplete: Use autocomplete for this option
        choices: The choices allowed by this command
        name: The name of the option. Defaults to the name of the argument

    """
    return SlashCommandOption(
        name=name,
        description=description,
        required=required,
        autocomplete=autocomplete,
        choices=choices or [],
        type=models.OptionType.ROLE,
    )


def slash_mentionable_option(
    description: str,
    required: bool = False,
    autocomplete: bool = False,
    choices: List[Union["SlashCommandChoice", dict]] | None = None,
    name: Optional[str] = None,
) -> Type[Union["Role", "BaseChannel", "User", "Member"]]:
    """
    Annotates an argument as a mentionable type slash command option.

    Args:
        description: The description of your option
        required: Is this option required?
        autocomplete: Use autocomplete for this option
        choices: The choices allowed by this command
        name: The name of the option. Defaults to the name of the argument

    """
    return SlashCommandOption(
        name=name,
        description=description,
        required=required,
        autocomplete=autocomplete,
        choices=choices or [],
        type=models.OptionType.MENTIONABLE,
    )


def slash_attachment_option(
    description: str,
    required: bool = False,
    name: Optional[str] = None,
) -> Type["Attachment"]:
    """
    Annotates an argument as an attachment type slash command option.

    Args:
        description: The description of your option
        required: Is this option required?
        name: The name of the option. Defaults to the name of the argument

    """
    return SlashCommandOption(
        name=name,
        description=description,
        required=required,
        type=models.OptionType.ATTACHMENT,
    )



"""Outlines the interaction between interactions and Discord's Gateway API."""

import asyncio
import logging
import sys
import time
import zlib
from asyncio import Task
from types import TracebackType
from typing import TypeVar, TYPE_CHECKING

from interactions.api import events
from interactions.client.const import MISSING, __api_version__
from interactions.client.utils.input_utils import FastJson
from interactions.client.utils.serializer import dict_filter_none
from interactions.models.discord.enums import Status
from interactions.models.discord.enums import WebSocketOPCode as OPCODE
from interactions.models.discord.snowflake import to_snowflake
from interactions.models.internal.cooldowns import CooldownSystem
from .websocket import WebsocketClient

if TYPE_CHECKING:
    from .state import ConnectionState
    from interactions.models.discord.snowflake import Snowflake_Type

__all__ = ("GatewayClient",)


SELF = TypeVar("SELF", bound="WebsocketClient")


class GatewayRateLimit:
    def __init__(self) -> None:
        self.lock = asyncio.Lock()
        # docs state 120 calls per 60 seconds, this is set conservatively to 110 per 60 seconds.
        rate = 110
        interval = 60
        self.cooldown_system = CooldownSystem(1, interval / rate)
        # hacky way to throttle how frequently we send messages to the gateway

    async def rate_limit(self) -> None:
        async with self.lock:
            while not self.cooldown_system.acquire_token():
                await asyncio.sleep(self.cooldown_system.get_cooldown_time())


class GatewayClient(WebsocketClient):
    """
    Abstraction over one gateway connection.

    Multiple `WebsocketClient` instances can be used to implement same-process sharding.

    Attributes:
        sequence: The sequence of this connection
        session_id: The session ID of this connection

    """

    def __init__(self, state: "ConnectionState", shard: tuple[int, int]) -> None:
        super().__init__(state)

        self.shard = shard

        self.chunk_cache = {}

        self._trace = []
        self.sequence = None
        self.session_id = None

        self.ws_url = state.gateway_url
        self.ws_resume_url = MISSING

        # This lock needs to be held to send something over the gateway, but is also held when
        # reconnecting. That way there's no race conditions between sending and reconnecting.
        self._race_lock = asyncio.Lock()
        # Then this event is used so that receive() can wait for the reconnecting to complete.
        self._closed = asyncio.Event()

        self._keep_alive = None
        self._kill_bee_gees = asyncio.Event()
        self._last_heartbeat = 0
        self._acknowledged = asyncio.Event()
        self._acknowledged.set()  # Initialize it as set

        self._ready = asyncio.Event()
        self._close_gateway = asyncio.Event()

        # Sanity check, it is extremely important that an instance isn't reused.
        self._entered = False

    async def __aenter__(self: SELF) -> SELF:
        if self._entered:
            raise RuntimeError("An instance of 'WebsocketClient' cannot be re-used!")

        self._entered = True
        self._zlib = zlib.decompressobj()

        self.ws = await self.state.client.http.websocket_connect(self.state.gateway_url)

        hello = await self.receive(force=True)
        self.heartbeat_interval = hello["d"]["heartbeat_interval"] / 1000
        self._closed.set()

        self._keep_alive = asyncio.create_task(self.run_bee_gees())

        await self._identify()

        return self

    async def __aexit__(
        self,
        exc_type: type[BaseException] | None,
        exc_val: BaseException | None,
        traceback: TracebackType | None,
    ) -> None:
        # Technically should not be possible in any way, but might as well be safe worst-case.
        self._close_gateway.set()

        try:
            if self._keep_alive is not None:
                self._kill_bee_gees.set()
                try:
                    # Even if we get cancelled that is fine, because then the keep-alive
                    # handler will also be cancelled since we're waiting on it.
                    await self._keep_alive  # Wait for the keep-alive handler to finish
                finally:
                    self._keep_alive = None
        finally:
            if self.ws is not None:
                # We could be cancelled here, it is extremely important that we close the
                # WebSocket either way, hence the try/except.
                try:
                    await self.ws.close(code=1000)
                finally:
                    self.ws = None

    async def run(self) -> None:
        """Start receiving events from the websocket."""
        while True:
            if self._stopping is None:
                self._stopping = asyncio.create_task(self._close_gateway.wait())
            receiving = asyncio.create_task(self.receive())
            done, _ = await asyncio.wait({self._stopping, receiving}, return_when=asyncio.FIRST_COMPLETED)

            if receiving in done:
                # Note that we check for a received message first, because if both completed at
                # the same time, we don't want to discard that message.
                msg = await receiving
            else:
                # This has to be the stopping task, which we join into the current task (even
                # though that doesn't give any meaningful value in the return).
                await self._stopping
                receiving.cancel()
                return

            op = msg.get("op")
            data = msg.get("d")
            seq = msg.get("s")
            event = msg.get("t")

            if seq:
                self.sequence = seq

            if op == OPCODE.DISPATCH:
                _ = asyncio.create_task(self.dispatch_event(data, seq, event))  # noqa: RUF006
                continue

            # This may try to reconnect the connection so it is best to wait
            # for it to complete before receiving more - that way there's less
            # possible race conditions to consider.
            await self.dispatch_opcode(data, op)

    async def dispatch_opcode(self, data, op: OPCODE) -> None:
        match op:
            case OPCODE.HEARTBEAT:
                self.state.wrapped_logger(logging.DEBUG, "❤ Received heartbeat request from gateway")
                return await self.send_heartbeat()

            case OPCODE.HEARTBEAT_ACK:
                self._latency.append(time.perf_counter() - self._last_heartbeat)

                if self._last_heartbeat != 0 and self._latency[-1] >= 15:
                    self.state.wrapped_logger(
                        logging.WARNING,
                        f"❤ High Latency! shard ID {self.shard[0]} heartbeat took {self._latency[-1]:.1f}s to be acknowledged!",
                    )
                else:
                    self.state.wrapped_logger(logging.DEBUG, "❤ Received heartbeat acknowledgement from gateway")

                return self._acknowledged.set()

            case OPCODE.RECONNECT:
                self.state.wrapped_logger(logging.DEBUG, "Gateway requested reconnect. Reconnecting...")
                return await self.reconnect(resume=True, url=self.ws_resume_url)

            case OPCODE.INVALIDATE_SESSION:
                self.state.wrapped_logger(logging.WARNING, "Gateway invalidated session. Reconnecting...")
                return await self.reconnect()

            case _:
                return self.state.wrapped_logger(logging.DEBUG, f"Unhandled OPCODE: {op} = {OPCODE(op).name}")

    async def dispatch_event(self, data, seq, event) -> None:
        match event:
            case "READY":
                self._ready.set()
                self._trace = data.get("_trace", [])
                self.sequence = seq
                self.session_id = data["session_id"]
                self.ws_resume_url = (
                    f"{data['resume_gateway_url']}?encoding=json&v={__api_version__}&compress=zlib-stream"
                )
                self.state.wrapped_logger(logging.INFO, "Gateway connection established")
                self.state.wrapped_logger(logging.DEBUG, f"Session ID: {self.session_id} Trace: {self._trace}")
                return self.state.client.dispatch(events.WebsocketReady(data))

            case "RESUMED":
                self.state._shard_ready.set()
                self.state.wrapped_logger(
                    logging.INFO, f"Successfully resumed connection! Session_ID: {self.session_id}"
                )
                self.state.client.dispatch(events.Resume())
                return None

            case "GUILD_MEMBERS_CHUNK":
                _ = asyncio.create_task(self._process_member_chunk(data.copy()))

            case _:
                # the above events are "special", and are handled by the gateway itself, the rest can be dispatched
                event_name = f"raw_{event.lower()}"
                if processor := self.state.client.processors.get(event_name):
                    try:
                        _ = asyncio.create_task(  # noqa: RUF006
                            processor(events.RawGatewayEvent(data.copy(), override_name=event_name))
                        )
                    except Exception as ex:
                        self.state.wrapped_logger(
                            logging.ERROR, f"Failed to run event processor for {event_name}: {ex}"
                        )
                else:
                    self.state.wrapped_logger(logging.DEBUG, f"No processor for `{event_name}`")

        self.state.client.dispatch(events.RawGatewayEvent(data.copy(), override_name="raw_gateway_event"))
        self.state.client.dispatch(events.RawGatewayEvent(data.copy(), override_name=f"raw_{event.lower()}"))

    def close(self) -> None:
        """Shutdown the websocket connection."""
        self._close_gateway.set()

    async def _identify(self) -> None:
        """Send an identify payload to the gateway."""
        if self.ws is None:
            raise RuntimeError
        payload = {
            "op": OPCODE.IDENTIFY,
            "d": {
                "token": self.state.client.http.token,
                "intents": self.state.intents,
                "shard": self.shard,
                "large_threshold": 250,
                "properties": {
                    "os": sys.platform,
                    "browser": "interactions",
                    "device": "interactions",
                },
                "presence": self.state.presence,
            },
            "compress": True,
        }

        serialized = FastJson.dumps(payload)
        await self.ws.send_str(serialized)

        self.state.wrapped_logger(
            logging.DEBUG, f"Identification payload sent to gateway, requesting intents: {self.state.intents}"
        )

    async def reconnect(self, *, resume: bool = False, code: int = 1012, url: str | None = None) -> None:
        self.state.clear_ready()
        self._ready.clear()
        await super().reconnect(resume=resume, code=code, url=url)

    async def _resume_connection(self) -> None:
        """Send a resume payload to the gateway."""
        if self.ws is None:
            raise RuntimeError

        payload = {
            "op": OPCODE.RESUME,
            "d": {
                "token": self.state.client.http.token,
                "seq": self.sequence,
                "session_id": self.session_id,
            },
        }

        serialized = FastJson.dumps(payload)
        await self.ws.send_str(serialized)

        self.state.wrapped_logger(logging.DEBUG, f"Resume payload sent to gateway, session ID: {self.session_id}")

    async def send_heartbeat(self) -> None:
        await self.send_json({"op": OPCODE.HEARTBEAT, "d": self.sequence}, bypass=True)
        self.state.wrapped_logger(logging.DEBUG, "❤ Gateway is sending a Heartbeat")

    async def change_presence(self, activity=None, status: Status = Status.ONLINE, since=None) -> None:
        """Update the bot's presence status."""
        await self.send_json(
            {
                "op": OPCODE.PRESENCE,
                "d": dict_filter_none(
                    {
                        "since": int(since or time.time() * 1000),
                        "activities": [activity] if activity else [],
                        "status": status,
                        "afk": False,
                    }
                ),
            }
        )

    async def request_member_chunks(
        self,
        guild_id: "Snowflake_Type",
        query="",
        *,
        limit,
        user_ids=None,
        presences=False,
        nonce=None,
    ) -> None:
        payload = {
            "op": OPCODE.REQUEST_MEMBERS,
            "d": dict_filter_none(
                {
                    "guild_id": guild_id,
                    "presences": presences,
                    "limit": limit,
                    "nonce": nonce,
                    "user_ids": user_ids,
                    "query": query,
                }
            ),
        }
        await self.send_json(payload)

    async def _process_member_chunk(self, chunk: dict) -> Task[None]:
        if guild := self.state.client.cache.get_guild(to_snowflake(chunk.get("guild_id"))):
            return asyncio.create_task(guild.process_member_chunk(chunk))
        raise ValueError(f"No guild exists for {chunk.get('guild_id')}")

    async def voice_state_update(
        self,
        guild_id: "Snowflake_Type",
        channel_id: "Snowflake_Type",
        muted: bool = False,
        deafened: bool = False,
    ) -> None:
        """Update the bot's voice state."""
        payload = {
            "op": OPCODE.VOICE_STATE,
            "d": {
                "guild_id": guild_id,
                "channel_id": channel_id,
                "self_mute": muted,
                "self_deaf": deafened,
            },
        }
        await self.send_json(payload)



from typing import TYPE_CHECKING, Any, List, Optional

import discord_typings

from interactions.client.const import MISSING
from ..route import Route

__all__ = ("StickerRequests",)


if TYPE_CHECKING:
    from interactions.models.discord.snowflake import Snowflake_Type
    from interactions import UPLOADABLE_TYPE


class StickerRequests:
    request: Any

    async def get_sticker(self, sticker_id: "Snowflake_Type") -> discord_typings.StickerData:
        """
        Get a specific sticker.

        Args:
            sticker_id: The id of the sticker

        Returns:
            Sticker or None

        """
        return await self.request(Route("GET", "/stickers/{sticker_id}", sticker_id=sticker_id))

    async def list_nitro_sticker_packs(self) -> List[discord_typings.StickerPackData]:
        """
        Gets the list of sticker packs available to Nitro subscribers.

        Returns:
            List of sticker packs

        """
        return await self.request(Route("GET", "/sticker-packs"))

    async def list_guild_stickers(self, guild_id: "Snowflake_Type") -> List[discord_typings.StickerData]:
        """
        Get the stickers for a guild.

        Args:
            guild_id: The guild to get stickers from

        Returns:
            List of Stickers or None

        """
        return await self.request(Route("GET", "/guilds/{guild_id}/stickers", guild_id=guild_id))

    async def get_guild_sticker(
        self, guild_id: "Snowflake_Type", sticker_id: "Snowflake_Type"
    ) -> discord_typings.StickerData:
        """
        Get a sticker from a guild.

        Args:
            guild_id: The guild to get stickers from
            sticker_id: The sticker to get from the guild

        Returns:
            Sticker or None

        """
        return await self.request(
            Route("GET", "/guilds/{guild_id}/stickers/{sticker_id}", guild_id=guild_id, sticker_id=sticker_id)
        )

    async def create_guild_sticker(
        self,
        payload: dict,
        guild_id: "Snowflake_Type",
        file: "UPLOADABLE_TYPE",
        reason: Optional[str] = MISSING,
    ) -> discord_typings.StickerData:
        """
        Create a new sticker for the guild. Requires the MANAGE_EMOJIS_AND_STICKERS permission.

        Args:
            payload: the payload to send.
            guild_id: The guild to create sticker at.
            file: the image to use for the sticker
            reason: The reason for this action.

        Returns:
            The new sticker data on success.

        """
        return await self.request(
            Route("POST", "/guilds/{guild_id}/stickers", guild_id=guild_id),
            payload=payload,
            files=[file],
            reason=reason,
        )

    async def modify_guild_sticker(
        self,
        payload: dict,
        guild_id: "Snowflake_Type",
        sticker_id: "Snowflake_Type",
        reason: Optional[str] = MISSING,
    ) -> discord_typings.StickerData:
        """
        Modify the given sticker. Requires the MANAGE_EMOJIS_AND_STICKERS permission.

        Args:
            payload: the payload to send.
            guild_id: The guild of the target sticker.
            sticker_id:  The sticker to modify.
            reason: The reason for this action.

        Returns:
            The updated sticker data on success.

        """
        return await self.request(
            Route("PATCH", "/guilds/{guild_id}/stickers/{sticker_id}", guild_id=guild_id, sticker_id=sticker_id),
            payload=payload,
            reason=reason,
        )

    async def delete_guild_sticker(
        self,
        guild_id: "Snowflake_Type",
        sticker_id: "Snowflake_Type",
        reason: Optional[str] = MISSING,
    ) -> None:
        """
        Delete the given sticker. Requires the MANAGE_EMOJIS_AND_STICKERS permission.

        Args:
            guild_id: The guild of the target sticker.
            sticker_id:  The sticker to delete.
            reason: The reason for this action.

        Returns:
            Returns 204 No Content on success.

        """
        return await self.request(
            Route("DELETE", "/guilds/{guild_id}/stickers/{sticker_id}", guild_id=guild_id, sticker_id=sticker_id),
            reason=reason,
        )



import copy
from typing import TYPE_CHECKING

import interactions.api.events as events
from interactions.models import to_snowflake, BaseMessage
from ._template import EventMixinTemplate, Processor

if TYPE_CHECKING:
    from interactions.api.events import RawGatewayEvent

__all__ = ("MessageEvents",)


class MessageEvents(EventMixinTemplate):
    @Processor.define()
    async def _on_raw_message_create(self, event: "RawGatewayEvent") -> None:
        """
        Automatically convert MESSAGE_CREATE event data to the object.

        Args:
            event: raw message event

        """
        msg = self.cache.place_message_data(event.data)
        if not msg._guild_id and event.data.get("guild_id"):
            msg._guild_id = event.data["guild_id"]

        if msg._guild_id and not msg.guild:
            await self.cache.fetch_guild(msg._guild_id)

        if not msg.channel:
            await self.cache.fetch_channel(to_snowflake(msg._channel_id))

        self.dispatch(events.MessageCreate(msg))

    @Processor.define()
    async def _on_raw_message_delete(self, event: "RawGatewayEvent") -> None:
        """
        Process raw deletions and dispatch a processed deletion event.

        Args:
            event: raw message deletion event

        """
        message = self.cache.get_message(
            event.data.get("channel_id"),
            event.data.get("id"),
        )

        if not message:
            message = BaseMessage.from_dict(event.data, self)
        self.cache.delete_message(event.data["channel_id"], event.data["id"])
        self.logger.debug(f"Dispatching Event: {event.resolved_name}")
        self.dispatch(events.MessageDelete(message))

    @Processor.define()
    async def _on_raw_message_update(self, event: "RawGatewayEvent") -> None:
        """
        Process raw message update event and dispatch a processed update event.

        Args:
            event: raw message update event

        """
        # a copy is made because the cache will update the original object in memory
        before = copy.copy(self.cache.get_message(event.data.get("channel_id"), event.data.get("id")))
        after = self.cache.place_message_data(event.data)
        self.dispatch(events.MessageUpdate(before=before, after=after))

    @Processor.define()
    async def _on_raw_message_delete_bulk(self, event: "RawGatewayEvent") -> None:
        """
        Process raw bulk message deletion event and dispatch a processed bulk deletion event.

        Args:
            event: raw bulk message deletion event

        """
        self.dispatch(
            events.MessageDeleteBulk(
                event.data.get("guild_id", None),
                event.data.get("channel_id"),
                event.data.get("ids"),
            )
        )

    @Processor.define()
    async def _on_raw_message_poll_vote_add(self, event: "RawGatewayEvent") -> None:
        """
        Process raw message poll vote add event and dispatch a processed poll vote add event.

        Args:
            event: raw poll vote add event

        """
        self.dispatch(
            events.MessagePollVoteAdd(
                event.data.get("guild_id", None),
                event.data["channel_id"],
                event.data["message_id"],
                event.data["user_id"],
                event.data["option"],
            )
        )

    @Processor.define()
    async def _on_raw_message_poll_vote_remove(self, event: "RawGatewayEvent") -> None:
        """
        Process raw message poll vote remove event and dispatch a processed poll vote remove event.

        Args:
            event: raw poll vote remove event

        """
        self.dispatch(
            events.MessagePollVoteRemove(
                event.data.get("guild_id", None),
                event.data["channel_id"],
                event.data["message_id"],
                event.data["user_id"],
                event.data["option"],
            )
        )



from .channel_events import ChannelEvents
from .guild_events import GuildEvents
from .integrations import IntegrationEvents
from .member_events import MemberEvents
from .message_events import MessageEvents
from .reaction_events import ReactionEvents
from .role_events import RoleEvents
from .scheduled_events import ScheduledEvents
from .stage_events import StageEvents
from .thread_events import ThreadEvents
from .user_events import UserEvents
from .voice_events import VoiceEvents
from ._template import Processor
from .auto_mod import AutoModEvents
from .entitlement_events import EntitlementEvents

__all__ = (
    "ChannelEvents",
    "GuildEvents",
    "IntegrationEvents",
    "MemberEvents",
    "MessageEvents",
    "ReactionEvents",
    "RoleEvents",
    "ScheduledEvents",
    "StageEvents",
    "ThreadEvents",
    "UserEvents",
    "VoiceEvents",
    "Processor",
    "AutoModEvents",
    "EntitlementEvents",
)



import uuid
from enum import IntEnum
from typing import Union, Optional, Any, TypeVar

import discord_typings

from interactions.client.const import MISSING
from interactions.client.mixins.serialization import DictSerializationMixin
from interactions.client.utils import dict_filter
from interactions.models.discord.components import ComponentType
from interactions.models.internal.application_commands import CallbackType

__all__ = ("InputText", "Modal", "ParagraphText", "ShortText", "TextStyles")

T = TypeVar("T", bound="InputText")


class TextStyles(IntEnum):
    SHORT = 1
    PARAGRAPH = 2


class InputText(DictSerializationMixin):
    def __init__(
        self,
        *,
        label: str,
        style: Union[TextStyles, int],
        custom_id: Optional[str] = MISSING,
        placeholder: Optional[str] = MISSING,
        value: Optional[str] = MISSING,
        required: bool = True,
        min_length: Optional[int] = MISSING,
        max_length: Optional[int] = MISSING,
    ) -> None:
        self.label = label
        self.style = style
        self.custom_id = custom_id or str(uuid.uuid4())
        self.placeholder = placeholder
        self.value = value
        self.required = required
        self.min_length = min_length
        self.max_length = max_length

        self.type = ComponentType.INPUT_TEXT

    def to_dict(
        self,
    ) -> dict[str, int | str | bool]:  # I couldn't find a discord_typings object for this
        return dict_filter(
            {
                "type": self.type,
                "label": self.label,
                "style": self.style,
                "custom_id": self.custom_id,
                "placeholder": self.placeholder,
                "value": self.value,
                "required": self.required,
                "min_length": self.min_length,
                "max_length": self.max_length,
            }
        )

    @classmethod
    def from_dict(cls, data: dict[str, Any]) -> T:
        if data["style"] == TextStyles.SHORT:
            cls = ShortText
        elif data["style"] == TextStyles.PARAGRAPH:
            cls = ParagraphText

        return cls(
            label=data["label"],
            custom_id=data["custom_id"],
            placeholder=data["placeholder"],
            value=data["value"],
            required=data["required"],
            min_length=data["min_length"],
            max_length=data["max_length"],
        )


class ShortText(InputText):
    def __init__(
        self,
        *,
        label: str,
        custom_id: Optional[str] = MISSING,
        placeholder: Optional[str] = MISSING,
        value: Optional[str] = MISSING,
        required: bool = True,
        min_length: Optional[int] = MISSING,
        max_length: Optional[int] = MISSING,
    ) -> None:
        super().__init__(
            style=TextStyles.SHORT,
            label=label,
            custom_id=custom_id,
            placeholder=placeholder,
            value=value,
            required=required,
            min_length=min_length,
            max_length=max_length,
        )


class ParagraphText(InputText):
    def __init__(
        self,
        *,
        label: str,
        custom_id: Optional[str] = MISSING,
        placeholder: Optional[str] = MISSING,
        value: Optional[str] = MISSING,
        required: bool = True,
        min_length: Optional[int] = MISSING,
        max_length: Optional[int] = MISSING,
    ) -> None:
        super().__init__(
            style=TextStyles.PARAGRAPH,
            label=label,
            custom_id=custom_id,
            placeholder=placeholder,
            value=value,
            required=required,
            min_length=min_length,
            max_length=max_length,
        )


class Modal:
    def __init__(
        self,
        *components: InputText,
        title: str,
        custom_id: Optional[str] = None,
    ) -> None:
        self.title: str = title
        self.components: list[InputText] = list(components)
        self.custom_id: str = custom_id or str(uuid.uuid4())

        self.type = CallbackType.MODAL

    def to_dict(self) -> discord_typings.ModalInteractionData:
        return {
            "type": self.type,
            "data": {
                "title": self.title,
                "custom_id": self.custom_id,
                "components": [
                    {
                        "type": ComponentType.ACTION_ROW,
                        "components": [c.to_dict() if hasattr(c, "to_dict") else c],
                    }
                    for c in self.components
                ],
            },
        }

    def add_components(self, *components: InputText) -> None:
        """
        Add components to the modal.

        Args:
            *components: The components to add.

        """
        if len(components) == 1 and isinstance(components[0], (list, tuple)):
            components = components[0]
        self.components.extend(components)



"""This file handles the interaction with discords http endpoints."""

import asyncio
import inspect
import os
import time
from logging import Logger
from typing import Any, cast, Callable
from urllib.parse import quote as _uriquote
from weakref import WeakValueDictionary

import aiohttp
import discord_typings
from aiohttp import BaseConnector, ClientSession, ClientWebSocketResponse, FormData, BasicAuth
from multidict import CIMultiDictProxy

import interactions.client.const as constants
from interactions import models
from interactions.api.http.http_requests import (
    BotRequests,
    ChannelRequests,
    EmojiRequests,
    EntitlementRequests,
    GuildRequests,
    InteractionRequests,
    MemberRequests,
    MessageRequests,
    ReactionRequests,
    StickerRequests,
    ThreadRequests,
    UserRequests,
    WebhookRequests,
    ScheduledEventsRequests,
)
from interactions.client.const import (
    MISSING,
    __py_version__,
    __repo_url__,
    __version__,
    __api_version__,
)
from interactions.client.errors import (
    DiscordError,
    Forbidden,
    GatewayNotFound,
    HTTPException,
    NotFound,
    LoginError,
)
from interactions.client.mixins.serialization import DictSerializationMixin
from interactions.client.utils.input_utils import response_decode, FastJson
from interactions.client.utils.serializer import dict_filter, get_file_mimetype
from interactions.models.discord.file import UPLOADABLE_TYPE
from .route import Route

__all__ = ("HTTPClient",)


class GlobalLock:
    def __init__(self) -> None:
        self._lock = asyncio.Lock()
        self.max_requests = 45
        self._calls = self.max_requests
        self._reset_time = 0

    @property
    def calls_remaining(self) -> int:
        """Returns the amount of calls remaining."""
        return self.max_requests - self._calls

    def reset_calls(self) -> None:
        """Resets the calls to the max amount."""
        self._calls = self.max_requests
        self._reset_time = time.perf_counter() + 1

    def set_reset_time(self, delta: float) -> None:
        """
        Sets the reset time to the current time + delta.

        To be called if a 429 is received.

        Args:
            delta: The time to wait before resetting the calls.

        """
        self._reset_time = time.perf_counter() + delta
        self._calls = 0

    async def wait(self) -> None:
        """Throttles calls to prevent hitting the global rate limit."""
        async with self._lock:
            if self._reset_time <= time.perf_counter():
                self.reset_calls()
            elif self._calls <= 0:
                await asyncio.sleep(self._reset_time - time.perf_counter())
                self.reset_calls()
        self._calls -= 1


class BucketLock:
    """Manages the rate limit for each bucket."""

    DEFAULT_LIMIT = 1
    DEFAULT_REMAINING = 1
    DEFAULT_DELTA = 0.0

    def __init__(self, header: CIMultiDictProxy | None = None) -> None:
        self._semaphore: asyncio.Semaphore | None = None
        if header is None:
            self.bucket_hash: str | None = None
            self.limit: int = self.DEFAULT_LIMIT
            self.remaining: int = self.DEFAULT_REMAINING
            self.delta: float = self.DEFAULT_DELTA
        else:
            self.ingest_ratelimit_header(header)

        self.logger = constants.get_logger()

        self._lock: asyncio.Lock = asyncio.Lock()

    def __repr__(self) -> str:
        return f"<BucketLock: {self.bucket_hash or 'Generic'}, limit: {self.limit}, remaining: {self.remaining}, delta: {self.delta}>"

    @property
    def locked(self) -> bool:
        """Returns whether the bucket is locked."""
        if self._lock.locked():
            return True
        return self._semaphore is not None and self._semaphore.locked()

    def ingest_ratelimit_header(self, header: CIMultiDictProxy) -> None:
        """
        Ingests the rate limit header.

        Args:
            header: The header to ingest, containing rate limit information.

        Updates the bucket_hash, limit, remaining, and delta attributes with the information from the header.

        """
        self.bucket_hash = header.get("x-ratelimit-bucket")
        self.limit = int(header.get("x-ratelimit-limit", self.DEFAULT_LIMIT))
        self.remaining = int(header.get("x-ratelimit-remaining", self.DEFAULT_REMAINING))
        self.delta = float(header.get("x-ratelimit-reset-after", self.DEFAULT_DELTA))

        if self.delta < 0.005 and self.remaining == 0:  # the delta value is so small that we can assume it's 0
            self.delta = self.DEFAULT_DELTA
            self.remaining = self.DEFAULT_REMAINING  # we can assume that we can make another request right away

        if self._semaphore is None or self._semaphore._value != self.limit:
            self._semaphore = asyncio.Semaphore(self.limit)

    async def acquire(self) -> None:
        """Acquires the semaphore."""
        if self._semaphore is None:
            return

        if self._lock.locked():
            self.logger.debug(f"Waiting for bucket {self.bucket_hash} to unlock.")
            async with self._lock:
                pass

        await self._semaphore.acquire()

    def release(self) -> None:
        """
        Releases the semaphore.

        Note: If the bucket has been locked with lock_for_duration, this will not release the lock.
        """
        if self._semaphore is None:
            return
        self._semaphore.release()

    async def lock_for_duration(self, duration: float, block: bool = False) -> None:
        """
        Locks the bucket for a given duration.

        Args:
            duration: The duration to lock the bucket for.
            block: Whether to block until the bucket is unlocked.

        Raises:
            RuntimeError: If the bucket is already locked.

        """
        if self._lock.locked():
            raise RuntimeError("Attempted to lock a bucket that is already locked.")

        async def _release() -> None:
            await asyncio.sleep(duration)
            self._lock.release()

        if block:
            await self._lock.acquire()
            await _release()
        else:
            await self._lock.acquire()
            _ = asyncio.create_task(_release())  # noqa: RUF006

    async def __aenter__(self) -> None:
        await self.acquire()

    async def __aexit__(self, *args) -> None:
        self.release()


class HTTPClient(
    BotRequests,
    ChannelRequests,
    EmojiRequests,
    EntitlementRequests,
    GuildRequests,
    InteractionRequests,
    MemberRequests,
    MessageRequests,
    ReactionRequests,
    StickerRequests,
    ThreadRequests,
    UserRequests,
    WebhookRequests,
    ScheduledEventsRequests,
):
    """A http client for sending requests to the Discord API."""

    def __init__(
        self,
        connector: BaseConnector | None = None,
        logger: Logger = MISSING,
        show_ratelimit_tracebacks: bool = False,
        proxy: tuple[str | None, BasicAuth | None] | None = None,
    ) -> None:
        self.connector: BaseConnector | None = connector
        self.__session: ClientSession | None = None
        self.token: str | None = None
        self.global_lock: GlobalLock = GlobalLock()
        self._max_attempts: int = 3

        self.ratelimit_locks: WeakValueDictionary[str, BucketLock] = WeakValueDictionary()
        self.show_ratelimit_traceback: bool = show_ratelimit_tracebacks
        self._endpoints = {}

        self.user_agent: str = (
            f"DiscordBot ({__repo_url__} {__version__} Python/{__py_version__}) aiohttp/{aiohttp.__version__}"
        )
        self.proxy: tuple[str | None, BasicAuth | None] | None = proxy
        self.__proxy_validated: bool = False

        if logger is MISSING:
            logger = constants.get_logger()
        self.logger = logger

    def get_ratelimit(self, route: Route) -> BucketLock:
        """
        Get a route's rate limit bucket.

        Args:
            route: The route to fetch the ratelimit bucket for

        Returns:
            The BucketLock object for this route

        """
        if bucket_hash := self._endpoints.get(route.rl_bucket):
            if lock := self.ratelimit_locks.get(bucket_hash):
                # if we have an active lock on this route, it'll still be in the cache
                # return that lock
                return lock
        # if no cached lock exists, return a new lock
        return BucketLock()

    def ingest_ratelimit(self, route: Route, header: CIMultiDictProxy, bucket_lock: BucketLock) -> None:
        """
        Ingests a ratelimit header from discord to determine ratelimit.

        Args:
            route: The route we're ingesting ratelimit for
            header: The rate limit header in question
            bucket_lock: The rate limit bucket for this route

        """
        bucket_lock.ingest_ratelimit_header(header)

        if bucket_lock.bucket_hash:
            # We only ever try and cache the bucket if the bucket hash has been set (ignores unlimited endpoints)
            self.logger.debug(f"Caching ingested rate limit data for: {bucket_lock.bucket_hash}")
            self._endpoints[route.rl_bucket] = bucket_lock.bucket_hash
            self.ratelimit_locks[bucket_lock.bucket_hash] = bucket_lock

    @staticmethod
    def _process_payload(
        payload: dict | list[dict] | None, files: UPLOADABLE_TYPE | list[UPLOADABLE_TYPE] | None
    ) -> dict | list[dict] | FormData | None:
        """
        Processes a payload into a format safe for discord. Converts the payload into FormData where required

        Args:
            payload: The payload of the request
            files: A list of any files to send

        Returns:
            Either a dictionary or multipart data form

        """
        if isinstance(payload, FormData):
            return payload
        if payload is None:
            return None

        if isinstance(payload, dict):
            payload = dict_filter(payload)

            for k, v in payload.items():
                if isinstance(v, DictSerializationMixin):
                    payload[k] = v.to_dict()
                if isinstance(v, (list, tuple, set)):
                    payload[k] = [i.to_dict() if isinstance(i, DictSerializationMixin) else i for i in v]

        else:
            payload = [dict_filter(x) if isinstance(x, dict) else x for x in payload]

        if files is None:
            return payload

        if files == []:
            payload["attachments"] = []
            return payload

        if not isinstance(files, list):
            files = (files,)

        attachments = []

        form_data = FormData(quote_fields=False)

        for index, file in enumerate(files):
            file_data = models.open_file(file).read()

            if isinstance(file, models.File):
                form_data.add_field(
                    f"files[{index}]",
                    file_data,
                    filename=file.file_name,
                    content_type=file.content_type or get_file_mimetype(file_data),
                )
                attachments.append({"id": index, "description": file.description, "filename": file.file_name})
            else:
                form_data.add_field(
                    f"files[{index}]",
                    file_data,
                    filename=file.split(os.sep)[-1],
                    content_type=get_file_mimetype(file_data),
                )
        if attachments:
            payload["attachments"] = attachments

        form_data.add_field("payload_json", FastJson.dumps(payload))
        return form_data

    async def request(  # noqa: C901
        self,
        route: Route,
        payload: list | dict | None = None,
        files: list[UPLOADABLE_TYPE] | None = None,
        reason: str | None = None,
        params: dict | None = None,
        **kwargs: dict,
    ) -> str | dict[str, Any] | None:
        """
        Make a request to discord.

        Args:
            route: The route to take
            payload: The payload for this request
            files: The files to send with this request
            reason: Attach a reason to this request, used for audit logs
            params: Query string parameters

        """
        # Assemble headers
        kwargs["headers"] = {"User-Agent": self.user_agent}
        if self.token:
            kwargs["headers"]["Authorization"] = f"Bot {self.token}"
        if reason:
            kwargs["headers"]["X-Audit-Log-Reason"] = _uriquote(reason, safe="/ ")

        if isinstance(payload, (list, dict)) and not files:
            kwargs["headers"]["Content-Type"] = "application/json"
        if isinstance(params, dict):
            kwargs["params"] = dict_filter(params)

        lock = self.get_ratelimit(route)
        # this gets a BucketLock for this route.
        # If this endpoint has been used before, it will get an existing ratelimit for the respective buckethash
        # otherwise a brand-new bucket lock will be returned

        for attempt in range(self._max_attempts):
            async with lock:
                try:
                    if self.__session.closed:
                        await self.login(cast(str, self.token))

                    processed_data = self._process_payload(payload, files)
                    if isinstance(processed_data, FormData):
                        kwargs["data"] = processed_data  # pyright: ignore
                    else:
                        kwargs["json"] = processed_data  # pyright: ignore
                    await self.global_lock.wait()

                    if self.proxy:
                        kwargs["proxy"] = self.proxy[0]
                        kwargs["proxy_auth"] = self.proxy[1]

                    async with self.__session.request(route.method, route.url, **kwargs) as response:
                        result = await response_decode(response)
                        self.ingest_ratelimit(route, response.headers, lock)

                        if response.status == 429:
                            # ratelimit exceeded
                            result = cast(dict[str, str], result)
                            if result.get("global", False):
                                # global ratelimit is reached
                                # if we get a global, that's pretty bad, this would usually happen if the user is hitting the api from 2 clients sharing a token
                                self.log_ratelimit(
                                    self.logger.warning,
                                    f"Bot has exceeded global ratelimit, locking REST API for {result['retry_after']} seconds",
                                )
                                self.global_lock.set_reset_time(float(result["retry_after"]))
                            elif result.get("message") == "The resource is being rate limited.":
                                # resource ratelimit is reached
                                self.log_ratelimit(
                                    self.logger.warning,
                                    f"{route.resolved_endpoint} The resource is being rate limited! "
                                    f"Reset in {result.get('retry_after')} seconds",
                                )
                                # lock this resource and wait for unlock
                                await lock.lock_for_duration(float(result["retry_after"]), block=True)
                            else:
                                # endpoint ratelimit is reached
                                # 429's are unfortunately unavoidable, but we can attempt to avoid them
                                # so long as these are infrequent we're doing well
                                self.log_ratelimit(
                                    self.logger.warning,
                                    f"{route.resolved_endpoint} Has exceeded its ratelimit ({lock.limit})! Reset in {lock.delta} seconds",
                                )
                                await lock.lock_for_duration(lock.delta, block=True)
                            continue
                        if lock.remaining == 0:
                            # Last call available in the bucket, lock until reset
                            self.log_ratelimit(
                                self.logger.debug,
                                f"{route.resolved_endpoint} Has exhausted its ratelimit ({lock.limit})! Locking route for {lock.delta} seconds",
                            )
                            await lock.lock_for_duration(
                                lock.delta
                            )  # lock this route, but continue processing the current response

                        elif response.status in {500, 502, 504}:
                            # Server issues, retry
                            self.logger.warning(
                                f"{route.resolved_endpoint} Received {response.status}... retrying in {1 + attempt * 2} seconds"
                            )
                            await asyncio.sleep(1 + attempt * 2)
                            continue

                        if not 300 > response.status >= 200:
                            await self._raise_exception(response, route, result)

                        self.logger.debug(
                            f"{route.resolved_endpoint} Received {response.status} :: [{lock.remaining}/{lock.limit} calls remaining]"
                        )
                        return result
                except OSError as e:
                    if attempt < self._max_attempts - 1 and e.errno in (54, 10054):
                        await asyncio.sleep(1 + attempt * 2)
                        continue
                    raise

    async def _raise_exception(self, response, route, result) -> None:
        self.logger.error(f"{route.method}::{route.url}: {response.status}")

        if response.status == 403:
            raise Forbidden(response, response_data=result, route=route)
        if response.status == 404:
            raise NotFound(response, response_data=result, route=route)
        if response.status >= 500:
            raise DiscordError(response, response_data=result, route=route)
        raise HTTPException(response, response_data=result, route=route)

    def log_ratelimit(self, log_func: Callable, message: str) -> None:
        """
        Logs a ratelimit message, optionally with a traceback if show_ratelimit_traceback is True

        Args:
            log_func: The logging function to use
            message: The message to log

        """
        if self.show_ratelimit_traceback:
            if frame := next(
                (frame for frame in inspect.stack() if constants.LIB_PATH not in frame.filename),
                None,
            ):
                frame_info = inspect.getframeinfo(frame[0])
                filename = os.path.relpath(frame_info.filename, os.getcwd())

                traceback = (
                    f"{filename}:{frame_info.lineno} in {frame_info.function}:: {frame_info.code_context[0].strip()}"
                )
                message = f"{message} | Caused By: {traceback}"

        log_func(message)

    async def request_cdn(self, url, asset) -> bytes:  # pyright: ignore [reportGeneralTypeIssues]
        self.logger.debug(f"{asset} requests {url} from CDN")
        async with self.__session.get(url) as response:
            if response.status == 200:
                return await response.read()
            await self._raise_exception(response, asset, await response_decode(response))

    async def login(self, token: str) -> dict[str, Any]:
        """
        "Login" to the gateway, basically validates the token and grabs user data.

        Args:
            token: the token to use

        Returns:
            The currently logged in bot's data

        """
        self.__session = ClientSession(
            connector=self.connector or aiohttp.TCPConnector(limit=self.global_lock.max_requests),
            json_serialize=FastJson.dumps,
        )
        if not self.__proxy_validated and self.proxy:
            try:
                self.logger.info(f"Validating Proxy @ {self.proxy[0]}")
                async with self.__session.get(
                    "http://icanhazip.com/", proxy=self.proxy[0], proxy_auth=self.proxy[1]
                ) as response:
                    if response.status != 200:
                        raise RuntimeError("Proxy configuration is invalid")
                    self.logger.info(f"Proxy Connected @ {(await response.text()).strip()}")
                    self.__proxy_validated = True
            except Exception as e:
                raise RuntimeError("Proxy configuration is invalid") from e

        self.token = token
        try:
            result = await self.request(Route("GET", "/users/@me"))
            return cast(dict[str, Any], result)
        except HTTPException as e:
            if e.status == 401:
                raise LoginError("An improper token was passed") from e
            raise

    async def close(self) -> None:
        """Close the session."""
        if self.__session and not self.__session.closed:
            await self.__session.close()

    async def get_gateway(self) -> str:
        """
        Gets the gateway url.

        Returns:
            The gateway url

        """
        try:
            result = await self.request(Route("GET", "/gateway"))
            result = cast(dict[str, Any], result)
        except HTTPException as exc:
            raise GatewayNotFound from exc
        return "{0}?encoding={1}&v={2}&compress=zlib-stream".format(result["url"], "json", __api_version__)

    async def get_gateway_bot(self) -> discord_typings.GetGatewayBotData:
        try:
            result = await self.request(Route("GET", "/gateway/bot"))
        except HTTPException as exc:
            raise GatewayNotFound from exc
        return cast(discord_typings.GetGatewayBotData, result)

    async def websocket_connect(self, url: str) -> ClientWebSocketResponse:
        """
        Connect to the websocket.

        Args:
            url: the url to connect to

        """
        return await self.__session.ws_connect(
            url,
            timeout=30,
            max_msg_size=0,
            autoclose=False,
            headers={"User-Agent": self.user_agent},
            compress=0,
            proxy=self.proxy[0] if self.proxy else None,
            proxy_auth=self.proxy[1] if self.proxy else None,
        )
